import io
import zlib
import sys
import base64
import traceback
import re
import inspect
import ctypes
import threading
import queue

try:
  from .checker import ASTChecker, CheckerConstants
except:
  from ratio1.code_cheker.checker import ASTChecker, CheckerConstants

__VER__ = '0.6.1'

UNALLOWED_DICT = {
  'import ': {
    'error': 'Imports are not allowed in plugin code ',
    'type': 'import',
  },

  'from ': {
    'error': 'Imports are not allowed in plugin code ',
    'type': 'import',
  },

  'globals': {
    'error': 'Global vars access is not allowed in plugin code ',
    'type': CheckerConstants.var,
  },

  '__builtins__': {
    'error': '__builtins__ access is not allowed in plugin code ',
    'type': CheckerConstants.var,
  },

  'locals': {
    'error': 'Local vars dict access is not allowed in plugin code ',
    'type': CheckerConstants.var,
  },

  'memoryview': {
    'error': 'Pointer handling is unsafe in plugin code ',
    'type': CheckerConstants.var,
  },

  'log': {
    'error': 'Logger object cannot be used directly in plugin code - please use API ',
    'type': CheckerConstants.attr,
  },

  'vars': {
    'error': 'Usage of `vars(obj)` is not allowed in plugin code ',
    'type': CheckerConstants.var,
  },

  'dir': {
    'error': 'Usage of `dir(obj)` is not allowed in plugin code ',
    'type': CheckerConstants.var,
  },

  'global_shmem': {
    'error': 'Usage of `global_shmem` is not allowed in plugin code ',
    'type': CheckerConstants.attr,
  },

  'plugins_shmem': {
    'error': 'Usage of `plugins_shmem` is not allowed in plugin code ',
    'type': CheckerConstants.attr,
  },

  'config_data': {
    'error': 'Usage of `config_data` is not allowed in plugin code ',
    'type': CheckerConstants.attr,
  },

  '_default_config': {
    'error': 'Usage of `_default_config` is not allowed in plugin code ',
    'type': CheckerConstants.attr,
  },

  '__traceback__': {
    'error': 'Usage of `__traceback__` as an attribute is not allowed in plugin code ',
    'type': CheckerConstants.attr,
  },

  '_upstream_config': {
    'error': 'Usage of `_upstream_config` is not allowed in plugin code ',
    'type': CheckerConstants.attr,
  },

  'exec': {
    'error': 'Usage of `exec()` is not allowed in plugin code ',
    'type': CheckerConstants.var,
  },

  'eval': {
    'error': 'Usage of `eval()` is not allowed in plugin code ',
    'type': CheckerConstants.var,
  },

  'getattr': {
    'error': 'Usage of `getattr()` is not allowed in plugin code ',
    'type': CheckerConstants.var,
  },

  'open': {
    'error': 'Usage of `open()` is not allowed in plugin code ',
    'type': CheckerConstants.var,
  }
}

RESULT_VARS = ['__result', '_result', 'result']


class CodeExecutionTimeoutError(Exception):
  pass


class BaseCodeChecker:
  """
  This class should be used either as a associated object for code checking or
  as a mixin for running code
  """

  def __init__(self):
    super(BaseCodeChecker, self).__init__()
    self.printed_lines = []
    self.__exec_code_lock = threading.Lock()
    return

  def __msg(self, m, color='d'):
    if hasattr(self, 'P'):
      self.P(m, color=color)
    elif hasattr(self, 'log'):
      self.log.P(m, color=color)
    else:
      print(m)
    return

  def _is_safe_import(self, code, safe_imports):
    if safe_imports is None:
      return False
    for imp in safe_imports:
      if imp in code:
        return True
    return False

  def _check_unsafe_code(self, code, safe_imports=None):
    checker = ASTChecker(UNALLOWED_DICT, safe_imports)
    errors = checker.validate(code)
    if len(errors) == 0:
      return None
    return errors

  # PUB

  def check_code_text(self, code, safe_imports=None):
    return self._check_unsafe_code(code, safe_imports=safe_imports)

  def str_to_base64(self, str, verbose=False, compress=False):
    l_i = len(str)
    l_c = -1
    b_str = bytes(str, 'utf-8')
    if compress:
      b_str = zlib.compress(b_str, level=9)
      l_c = sys.getsizeof(b_str)
    b_encoded = base64.b64encode(b_str)
    str_encoded = b_encoded.decode('utf-8')
    l_b64 = len(str_encoded)
    if verbose:
      self.__msg("Initial/Compress/B64: {}/{}/{}".format(
        l_i, l_c, l_b64), color='g'
      )
    return str_encoded

  def code_to_base64(self, code, verbose=False, compress=True, return_errors=False):
    if verbose:
      self.__msg("Processing:\n{}".format(code), color='y')
    errors = self._check_unsafe_code(code)
    if errors is not None:
      err_msg = "Cannot serialize code due to: '{}'".format(errors)
      self.__msg(err_msg, color='r')
      return None if not return_errors else (None, err_msg)
    self.__msg("Code checking succeeded", color='g')
    str_encoded = self.str_to_base64(code, verbose=verbose, compress=compress)
    return str_encoded if not return_errors else (str_encoded, None)

  def base64_to_code(self, b64code, decompress=True):
    decoded = None
    try:
      b_decoded = base64.b64decode(b64code)
      if decompress:
        b_decoded = zlib.decompress(b_decoded)
      s_decoded = b_decoded.decode('utf-8')
      decoded = s_decoded
    except:
      pass
    return decoded

  def prepare_b64code(self, str_b64code, check_for_result=True, result_vars=RESULT_VARS):
    errors, code = None, None
    code = self.base64_to_code(str_b64code)
    to_check_code = code
    if code is not None:
      if self._can_encapsulate_code_in_method(code):
        # we have a return statement in the code,
        # so we need to encapsulate the code in a function
        to_check_code = self._encapsulate_code_in_method(
          exec_code__code=code,
          exec_code__arguments=[]
        )
      # endif can encapsulate code in method
      errors = self._check_unsafe_code(to_check_code)
    if errors is None:
      if code is None:
        errors = 'Provided ascii data is not a valid base64 object'
      # endif no valid code provided
    # endif no errors
    return code, errors

  def _add_line_after_each_line(self, code, codeline='plugin.sleep(0.001)'):
    lines = code.splitlines()
    refactor = []
    has_loop = False
    for line in lines:
      rstripped = line.rstrip()
      stripped = line.lstrip()
      is_loop = stripped.startswith(('while ', 'for '))
      has_loop = has_loop or is_loop
      if is_loop and rstripped[-1] != ':':
        parts = line.split(':')
        if len(parts) > 0:
          line = parts[0] + ': ' + codeline + ';' + parts[1]
          has_loop = False  # loop solved
      elif has_loop and not is_loop:
        nspc = len(line) - len(stripped)
        spc = nspc * ' '
        refactor.append(spc + codeline)
        has_loop = False  # loop solved
      # endif
      refactor.append(line)
    str_refactor = '\n'.join(refactor)
    return str_refactor

  def _can_encapsulate_code_in_method(self, exec_code__code):
    return re.search(r'\breturn\b', exec_code__code) is not None

  def _encapsulate_code_in_method(self, exec_code__code, exec_code__arguments):
    for i in range(len(exec_code__arguments)):
      __var = exec_code__arguments[i]
      if isinstance(__var, tuple):
        exec_code__arguments[i] = f"{__var[0]}={__var[1]}"
    exec_code__arguments = ', '.join(exec_code__arguments)

    if re.search(r'\breturn\b', exec_code__code) is not None:
      # we have a return statement in the code,
      # so we need to encapsulate the code in a function
      # 1. indent the code
      exec_code__code = "\n".join(['  ' + l for l in exec_code__code.splitlines()])
      # 2. add the function definition
      exec_code__code = "{}\n{}".format(
        f"def __exec_code__({exec_code__arguments}):",
        exec_code__code,
      )
    return exec_code__code

  def custom_print(self, print_queue, *args, **kwargs):
    """
    Custom print function that will be used in the plugin code.
    """
    # redirect the print to the plugin cache
    outstream = io.StringIO()
    print(*args, file=outstream, **kwargs)
    printed_value = outstream.getvalue()
    print_queue.put(printed_value)
    # print to the console
    print(f'[CUST_CODE_PRINT]{printed_value}')
    return

  def execute_code(self, code, local_vars, output_queue, print_queue):
    exec_code__result_var = None
    exec_code__warnings = []

    local_vars['print'] = lambda *args, **kwargs: self.custom_print(print_queue, *args, **kwargs)
    try:
      # Execute the code
      with self.__exec_code_lock:
        exec(code, local_vars)

        # Capture result variables
        for _var in local_vars.get('exec_code__result_vars', []):
          if _var in local_vars:
            exec_code__result_var = local_vars[_var]
            break
        # endfor all result vars
      # endwith lock
      # Put the results in the queue
      output_queue.put({
        "result_var": exec_code__result_var,
        "warnings": exec_code__warnings,
        "error": None
      })
    except Exception as e:
      output_queue.put({
        "result_var": None,
        "warnings": exec_code__warnings,
        "error": traceback.format_exc()
      })

  def __code_exec_stop_thread(self, thread):
    """
    Stop the specified thread.
    Parameters
    ----------
    thread : threading.Thread
        The thread to stop.
    """
    tid = thread.ident
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), ctypes.py_object(SystemExit))
    if res == 0:
      raise ValueError("Invalid thread ID")
    elif res > 1:
      # If it modifies more than one thread, something went wrong, so revert it
      ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), 0)
      raise SystemError("PyThreadState_SetAsyncExc failed")

  def execute_code_with_timeout(self, code, timeout, local_vars=None):
    if local_vars is None:
      local_vars = {}

    # Queue to collect output
    output_queue = queue.Queue()
    print_queue = queue.Queue()

    # Create a separate thread for code execution
    thread = threading.Thread(target=self.execute_code, args=(code, local_vars, output_queue, print_queue))
    thread.daemon = True
    # process = multiprocessing.Process(target=self.execute_code, args=(code, local_vars, output_queue, print_queue))

    # Start the process
    # process.start()
    thread.start()

    # Wait for the process to complete or timeout
    # process.join(timeout)
    thread.join(timeout)

    # If process is still alive after timeout, terminate it
    if thread.is_alive():
      # process.terminate()
      # thread.join()
      # TODO: maybe still send partial results or prints?
      self.__code_exec_stop_thread(thread)

      return {
        "result_var": None,
        "warnings": [],
        "printed_lines": [],
        "error": f"Code execution took longer than {timeout} seconds."
      }
    # endif process is still alive

    printed_lines = []
    while not print_queue.empty():
      printed_lines.append(print_queue.get())
    # Get the output from the queue
    if not output_queue.empty():
      exec_result = output_queue.get()
      exec_result['printed_lines'] = printed_lines
      return exec_result

    return {
      "result_var": None,
      "warnings": [],
      "printed_lines": printed_lines,
      "error": "No result returned."
    }

  def exec_code(self, str_b64code, debug=False, result_vars=None, self_var=None, modify=True, return_printed=False, timeout=None):
    exec_code__result_vars = result_vars or RESULT_VARS
    exec_code__warnings = []
    exec_code__result_var = None

    # Prepare the code
    exec_code__code, exec_code__errors = self.prepare_b64code(str_b64code, result_vars=exec_code__result_vars)

    if exec_code__errors:
        self.__msg(f"Cannot execute remote code: {exec_code__errors}", color='r')
        return exec_code__result_var, exec_code__errors, exec_code__warnings

    # Optionally modify the code
    if modify:
        exec_code__code = self._add_line_after_each_line(code=exec_code__code)

    if debug:
        self.__msg(f"DEBUG EXEC: Executing:\n{exec_code__code}")

    # Add `self` to locals if specified
    local_vars = locals().copy()
    if self_var and isinstance(self_var, str) and len(self_var) > 3:
        local_vars[self_var] = self

    # Handle encapsulating the code in a method if needed
    if self._can_encapsulate_code_in_method(exec_code__code):
        exec_code__code = self._encapsulate_code_in_method(exec_code__code, exec_code__arguments=[self_var])
        exec_code__code = f"{exec_code__code}\nresult = __exec_code__({self_var})"

    # Prepare to capture printed output
    self.printed_lines = []
    local_vars['print'] = self.custom_print

    # Execute the code with a timeout
    exec_result = self.execute_code_with_timeout(exec_code__code, timeout, local_vars=local_vars)

    exec_code__result_var = exec_result.get("result_var")
    exec_code__errors = exec_result.get("error")
    exec_code__warnings.extend(exec_result.get("warnings"))

    # Collect results
    res = (exec_code__result_var, exec_code__errors, exec_code__warnings)

    if return_printed:
        res += (exec_result.get("printed_lines", []),)

    return res

  def _get_method_from_custom_code(
    self, 
    str_b64code, 
    debug=False, 
    result_vars=RESULT_VARS, 
    self_var=None, 
    modify=True, 
    method_arguments=[]
    ):
    exec_code__result_vars = result_vars
    exec_code__debug = debug
    exec_code__self_var = self_var
    exec_code__modify = modify
    exec_code__warnings = []
    exec_code__code, exec_code__errors = self.prepare_b64code(
      str_b64code,
      result_vars=exec_code__result_vars,
    )
    exec_code__result_var = None
    has_result = False
    if exec_code__errors is not None:
      self.__msg("Cannot execute remote code: {}".format(exec_code__errors), color='r')
      return exec_code__result_var, exec_code__errors, exec_code__warnings

    # code does not have any safety errors
    if exec_code__modify:
      exec_code__code = self._add_line_after_each_line(code=exec_code__code)
    if exec_code__debug:
      self.__msg("DEBUG EXEC: Executing: \n{}".format(exec_code__code))
    if exec_code__self_var is not None and isinstance(exec_code__self_var, str) and len(exec_code__self_var) > 3:
      locals()[exec_code__self_var] = self

    try:
      if self._can_encapsulate_code_in_method(exec_code__code):
        # we have a return statement in the code,
        # so we need to encapsulate the code in a function
        exec_code__code = self._encapsulate_code_in_method(
          exec_code__code=exec_code__code,
          exec_code__arguments=method_arguments
        )
        exec_code__code = "{}\n{}".format(
          exec_code__code,
          f"result = __exec_code__"
        )
      else:
        # in this case we want to have our code in a method
        # so we will break here
        exec_code__errors = ["Cannot encapsulate code in method. No return statement found."]
        return exec_code__result_var, exec_code__errors, exec_code__warnings
      # endif can encapsulate code in method
      if exec_code__debug:
        self.__msg("DEBUG EXEC: Encapsulated code: \n{}".format(exec_code__code))
      exec(exec_code__code) # now we execute the method definition code not the actual method
      if exec_code__debug:
        self.__msg("DEBUG EXEC: locals(): \n{}".format(locals()))
      for _var in exec_code__result_vars:
        if _var in locals():
          if exec_code__debug:
            self.__msg("DEBUG EXEC: Extracting var '{}' from {}".format(_var, locals()))
          exec_code__result_var = locals().get(_var)
          has_result = True
          break
      if not has_result:
        exec_code__warnings.append("No result variable is set. Possible options: {}".format(exec_code__result_vars))
    except Exception as e:
      exec_code__result_var = None
      if hasattr(self, 'log'):
        exec_code__errors = list(self.log.get_error_info())
        exec_code__errors.append(traceback.format_exc())
      else:
        exec_code__errors = str(e)
    # end try-except
    return exec_code__result_var, exec_code__errors, exec_code__warnings

  def method_to_base64(self, func, verbose=False):
    code = self.get_function_source_code(func)
    return self.code_to_base64(code, verbose=verbose)

  def get_function_source_code(self, func):
    """
    Get the source code of a function including the docstring and remove the indentation.

    Parameters
    ----------
    func : Callable
        The function.

    Returns
    -------
    str
        The source code of the function.
    """
    plain_code = inspect.getsourcelines(func)[0]
    plain_code = plain_code[1:]
    first_code_line = 0
    # ignore empty lines at the beginning, but keep them
    while plain_code[first_code_line].strip() == '':
      first_code_line += 1
    indent = len(plain_code[first_code_line]) - len(plain_code[first_code_line].lstrip())
    plain_code = '\n'.join([line.rstrip()[indent:] for line in plain_code])

    return plain_code
  
  def _get_method_data(self, method: callable):
    """
    Get the full signature and definition of a method.
    
    Parameters
    ----------
    
    method : callable
        The method to get the signature and definition for.
        
    Returns
    -------
    tuple
        A tuple containing the name, arguments and base64 code of the method.
    """
    import inspect

    name = method.__name__
    args = list(map(str, inspect.signature(method).parameters.values()))
    ## 
    first_arg = args[0]
    first_arg = first_arg.split(':')[0]
    first_arg = first_arg.split('=')[0]
    first_arg = first_arg.strip()
    if first_arg in ['self', 'cls', 'plugin']:
      args = args[1:]
    ##      
    source = self.get_function_source_code(method)
    base64_code = self.code_to_base64(source)

    return name, args, base64_code  


if __name__ == '__main__':
  
  def some_function(plugin : int, x):
    """
    A simple function that adds 1 to the input.

    Parameters
    ----------
    x : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    print(f"Called by: {plugin}")
    return x + 1
  
  
  class SomeClass:
    def __init__(self, eng : BaseCodeChecker, str_b64code, func_args):
      self.eng : BaseCodeChecker = eng
      self._func, _, _ = eng._get_method_from_custom_code(
        str_b64code=str_b64code,
        self_var='plugin',
        method_arguments=["plugin"] + func_args,
        debug=False
      )  
      return
    
    
    def my_func(self, a):
      res = self._func(x=a, plugin=self)
      return res
      

  checker = BaseCodeChecker()
  source_code = checker.get_function_source_code(some_function)
  print("some_function:\n" + source_code)
  
  name, args, b64code = checker._get_method_data(some_function)
  print(f"Method name: {name}")
  print(f"Method args: {args}")
  print(f"Method code: {b64code}")
  
  target = SomeClass(checker, str_b64code=b64code, func_args=args)
  
  print(some_function(x=9.1, plugin=None))
  print(target.my_func(9.1))
  
  
  
  
  
  
  


"""
ex01_part3_adv_filter.py
---------------------
This is a simple example of how to use the ratio1 SDK.

In this example:
  - we connect to the network
  - listen for heartbeats from ratio1 Edge Protocol edge nodes and print the CPU of each node.
  - listen for payloads from ratio1 Edge Protocol edge nodes and print the data of each payload.
"""
import json

from ratio1 import Session, Payload, PAYLOAD_DATA, HEARTBEAT_DATA


class MessageHandler:
  def __init__(self, signature_filter: str = None):
    """
    This class is used to handle the messages received from the edge nodes.
    
    In this class we are defining two callback methods:
      - on_heartbeat: this method is called when a heartbeat is received from an edge node.
      - on_data: this method is called when a payload is received from an edge node.
    """
    if isinstance(signature_filter, str):
      self.signature_filter = [signature_filter.upper()]
    elif isinstance(signature_filter, list):
      self.signature_filter = [sig.upper() for sig in signature_filter]
    else:
      raise ValueError("signature_filter must be a string or a list of strings")
    self.last_data = None # some variable to store the last data received for debugging purposes
    self.last_payload = None # some variable to store the last payload received for debugging purposes
    return
  
  def shorten_address(self, address):
    """
    This method is used to shorten the address of the edge node.
    """
    return address[:8] + "..." + address[-6:]
  
  def on_heartbeat(self, session: Session, node_addr: str, heartbeat: dict):
    """
    This method is called when a heartbeat is received from an edge node.
    
    Parameters
    ----------
    session : Session
        The session object that received the heartbeat.
        
    node_addr : str
        The address of the edge node that sent the heartbeat.
        
    heartbeat : dict
        The heartbeat received from the edge node.        
    """
    whitelist = heartbeat.get(HEARTBEAT_DATA.EE_WHITELIST, [])
    alias = heartbeat.get(HEARTBEAT_DATA.EE_ID)
    cpu = heartbeat.get(HEARTBEAT_DATA.CPU)
    # DELETE:
    allows_me = session.bc_engine.contains_current_address(whitelist)
    # END DELETE
    session.P(
      "{} <{}> {} has {} and allows:\n{}".format(
        alias, self.shorten_address(node_addr), 
        "allows me" if allows_me else "does NOT allow me",
        cpu, json.dumps(whitelist, indent=2)),
      color='b' if allows_me else 'r',
    )
    return

  def on_data(
    self,
    session: Session, 
    node_addr : str, 
    pipeline_name : str, 
    plugin_signature : str, 
    plugin_instance : str,  
    payload : Payload      
  ):
    """
    This method is called when a payload is received from an edge node.
    
    Parameters
    ----------
    
    session : Session
        The session object that received the payload.
        
    node_addr : str
        The address of the edge node that sent the payload.
        
    pipeline_name : str
        The name of the pipeline that sent the payload.
        
    plugin_signature : str
        The signature of the plugin that sent the payload.
        
    plugin_instance : str
        The instance of the plugin that sent the payload.
        
    payload : Payload
        The payload received from the edge node.      
    """
    addr = self.shorten_address(node_addr)
    
    if plugin_signature.upper() not in self.signature_filter:
      # we are not interested in this data but we still want to log it
      message = "Recv from <{}::{}::{}::{}>".format(
        addr, pipeline_name, plugin_signature, plugin_instance
      )
      color = 'dark'
    else:
      self.last_payload = payload # save the full payload for debugging purposes
      # we can also access the "payload path" that matches 
      # the node-alias, pipeline_name, plugin_signature, and plugin_instance
      path = payload.EE_PAYLOAD_PATH 
      # we extract the data from the payload to check for online nodes considering
      # that we are waiting for NET_MON_01 payloads      
      # now we do some low-level processing of the data
      network_map = payload.data.get(PAYLOAD_DATA.NETMON_CURRENT_NETWORK, {})
      all_nodes = list(network_map.keys())
      online_nodes = [
        n for n in all_nodes 
        if network_map[n][PAYLOAD_DATA.NETMON_STATUS_KEY] == PAYLOAD_DATA.NETMON_STATUS_ONLINE
      ]
      whitelists = {
        k : v.get(PAYLOAD_DATA.NETMON_WHITELIST, []) for k, v in network_map.items()
      }
      message = f"{path[0]} Reports {len(online_nodes)} online nodes of {len(all_nodes)} known nodes. Whitelists with {len(whitelists)} peers"
      color = 'g'
    session.P(message, color=color, show=True)  #, noprefix=True)
    return


if __name__ == '__main__':
  # create a naive message handler for network monitoring public messages
  filterer = MessageHandler("NET_MON_01")
  
  # create a session
  # the network credentials are read from the .env file automatically
  session = Session(
      on_heartbeat=filterer.on_heartbeat,
      on_payload=filterer.on_data,
      # silent=True,
      debug=True,
  )

  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   due to the fact that this is a example/tutorial and maybe we dont want to keep the pipeline
  #   active after the session is closed we use close_pipelines=True
  #   in production, you would not need this code as the script can close 
  #   after the pipeline will be sent 
  session.wait(
    seconds=20,            # we wait the session for 60 seconds
    close_pipelines=True,   # we close the pipelines after the session
    close_session=True,     # we close the session after the session
  )
  session.P("Main thread exiting...")
  
  netinfo = session.get_network_known_nodes()  
  print(f"Last NET_MON_01 received:\n{json.dumps(filterer.last_payload.data, indent=2)}")
  print(f"Payload size: {len(json.dumps(filterer.last_payload.data))} bytes")
  print(f"Supervisor <{netinfo.reporter}> '{netinfo.reporter_alias}' ({netinfo.nr_super} supervisors total) reports:\n{netinfo.report}")



"""
ex13_launch_repo_based_webapp.py
-------------------------------

This script demonstrates how to create and deploy a webapp based on a public repository with ngrok or cloudflare
as tunnelling engine.
The script will create a webapp based on a public repository and deploy it to the
target Ratio1 Edge Node.

If an ngrok edge label or a cloudflare token is provided, the webapp will be accessible via a permanent URL.
In case ngrok is used and no edge label is provided, a unique URL will be generated by the ngrok service.
The temporary URL will be retrieved and printed to the console.

Note: In the case of a temporary URL, if the node is restarted, a new URL will be generated.

"""
from os import environ
from ratio1 import Session


if __name__ == "__main__":
  session = Session()

  node = environ.get("EE_TARGET_NODE", "INSERT_YOUR_NODE_ADDRESS_HERE")

  session.wait_for_node(node)

  # This is optional and if not tunnel engine is provided, the default will be ngrok.
  tunnel_engine = environ.get("TUNNEL_ENGINE", "ngrok").lower()

  # In case of using ngrok this is optional!
  # If an ngrok edge label or a cloudflare token is provided, the webapp will be accessible via a permanent URL.
  # In case this is not provided by the user and ngrok is used, a unique URL will be generated by the ngrok service.
  # Warning! create_webapp method does not support multiple nodes.
  # If you want to deploy the webapp to multiple nodes, you can use the create_and_deploy_balanced_web_app method
  ngrok_edge_label = environ.get("NGROK_EDGE_LABEL", None)
  cloudflare_token = environ.get("CLOUDFLARE_TOKEN", None)

  # Defining the environment variables for the webapp.
  # Here we can pass anything we want to the webapp.
  env_vars = {
    'LOCAL_ADDRESS': '/edge_node/_local_cache/_data/local_info.json',
    "VAR_NAME_1": "value1",
    "VAR_NAME_2": "value2",
  }

  # Defining the webapp setup commands.
  setup_commands = [
    "npm install"
  ]

  # Defining the assets of the application.
  assets = {
    'operation': "clone",
    "url": "https://github.com/ratio1/demo-deploy-nodejs",
    # The below are only necessary for private repos
    "username": "<username>",
    "token": "<git_user_token>",
  }

  # Defining the webapp run commands.
  run_command = "npm start"

  # instance: PLUGIN_TYPES.CUSTOM_WEBAPI_01
  pipeline, instance = session.create_web_app(
    node=node,
    name="Ratio1_WebApp_tutorial",
    env_vars=env_vars,
    setup_commands=setup_commands,
    run_command=run_command,
    tunnel_engine=tunnel_engine,
    ngrok_edge_label=ngrok_edge_label,
    cloudflare_token=cloudflare_token,
    assets=assets,
  )

  url = pipeline.deploy()

  # print the url of the webapp
  if ngrok_edge_label is None and cloudflare_token is None:
    session.P(
      f"Webapp available at: {url}.\nNote that this url is temporary. "
      f"In case the node is restarted, a new url will be generated.\n"
      f"In case you see an ngrok gateway error please try again after a few seconds.",
      color='g', boxed=True, show=True
    )
  else:
    msg = f"Webapp deployed using {tunnel_engine}. "
    msg += "For the url check the settings of the provided ngrok edge label or cloudflare token."
    session.P(msg, color='g', boxed=True, show=True)
  # endif ngrok_edge_label is None and cloudflare_token is None

  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   in production, you would not need this code as the script can close after the pipeline will be sent
  session.run(
    wait=True,  # wait for the user to stop the execution
    close_pipelines=True  # when the user stops the execution, the remote edge-node pipelines will be closed
  )



import json
import os

from datetime import datetime

from ratio1 import Session, Payload


class MessageHandler:
  def __init__(self):
    self.notifications = []
    self.heartbeats = []
    self.payloads = []

  def on_heartbeat(self, session: Session, node_addr: str, heartbeat: dict):
    self.heartbeats.append(heartbeat)
    return

  def on_payload(
      self,
      session: Session,
      node_addr: str,
      pipeline_name: str,
      plugin_signature: str,
      plugin_instance: str,
      payload: Payload
  ):
    self.payloads.append(payload.data)
    return

  def on_notification(self, session, address, notification):
    self.notifications.append(notification.data)
    return


if __name__ == '__main__':
  COLLECTION_PERIOD_SECONDS = 3600
  # create a naive message handler
  sniffer = MessageHandler()

  start_date_iso = datetime.now().isoformat()
  # create a session
  session = Session(
    on_heartbeat=sniffer.on_heartbeat,
    on_payload=sniffer.on_payload,
    on_notification=sniffer.on_notification,
    verbosity=3,
  )

  session.wait(seconds=COLLECTION_PERIOD_SECONDS, close_session=True)
  session.P("Session finished successfully!", color='g')

  folder_name = f"messages_{start_date_iso}_{COLLECTION_PERIOD_SECONDS}"
  os.makedirs(folder_name, exist_ok = True)

  session.P("Script run successfully!")
  session.close()



import logging
import os
import re
import sys
import warnings
from datetime import timezone

from . import utils

if sys.version_info >= (3, 9):
  import zoneinfo  # pragma: no cover
else:
  from backports import zoneinfo  # pragma: no cover

_cache_tz = None
_cache_tz_name = None

log = logging.getLogger("tzlocal")


def _get_localzone_name(_root="/"):
  """Tries to find the local timezone configuration.

  This method finds the timezone name, if it can, or it returns None.

  The parameter _root makes the function look for files like /etc/localtime
  beneath the _root directory. This is primarily used by the tests.
  In normal usage you call the function without parameters."""

  # First try the ENV setting.
  tzenv = utils._tz_name_from_env()
  if tzenv:
    return tzenv

  # Are we under Termux on Android?
  if os.path.exists(os.path.join(_root, "system/bin/getprop")):
    log.debug("This looks like Termux")

    import subprocess

    try:
      androidtz = (
          subprocess.check_output(["getprop", "persist.sys.timezone"])
          .strip()
          .decode()
      )
      return androidtz
    except (OSError, subprocess.CalledProcessError):
      # proot environment or failed to getprop
      log.debug("It's not termux?")
      pass

  # Now look for distribution specific configuration files
  # that contain the timezone name.

  # Stick all of them in a dict, to compare later.
  found_configs = {}

  for configfile in ("etc/timezone", "var/db/zoneinfo"):
    tzpath = os.path.join(_root, configfile)
    try:
      with open(tzpath) as tzfile:
        data = tzfile.read()
        log.debug(f"{tzpath} found, contents:\n {data}")

        etctz = data.strip("/ \t\r\n")
        if not etctz:
          # Empty file, skip
          continue
        for etctz in etctz.splitlines():
          # Get rid of host definitions and comments:
          if " " in etctz:
            etctz, dummy = etctz.split(" ", 1)
          if "#" in etctz:
            etctz, dummy = etctz.split("#", 1)
          if not etctz:
            continue

          found_configs[tzpath] = etctz.replace(" ", "_")

    except (OSError, UnicodeDecodeError):
      # File doesn't exist or is a directory, or it's a binary file.
      continue

  # CentOS has a ZONE setting in /etc/sysconfig/clock,
  # OpenSUSE has a TIMEZONE setting in /etc/sysconfig/clock and
  # Gentoo has a TIMEZONE setting in /etc/conf.d/clock
  # We look through these files for a timezone:

  zone_re = re.compile(r"\s*ZONE\s*=\s*\"")
  timezone_re = re.compile(r"\s*TIMEZONE\s*=\s*\"")
  end_re = re.compile('"')

  for filename in ("etc/sysconfig/clock", "etc/conf.d/clock"):
    tzpath = os.path.join(_root, filename)
    try:
      with open(tzpath, "rt") as tzfile:
        data = tzfile.readlines()
        log.debug(f"{tzpath} found, contents:\n {data}")

      for line in data:
        # Look for the ZONE= setting.
        match = zone_re.match(line)
        if match is None:
          # No ZONE= setting. Look for the TIMEZONE= setting.
          match = timezone_re.match(line)
        if match is not None:
          # Some setting existed
          line = line[match.end():]
          etctz = line[: end_re.search(line).start()]

          # We found a timezone
          found_configs[tzpath] = etctz.replace(" ", "_")

    except (OSError, UnicodeDecodeError):
      # UnicodeDecode handles when clock is symlink to /etc/localtime
      continue

  # systemd distributions use symlinks that include the zone name,
  # see manpage of localtime(5) and timedatectl(1)
  tzpath = os.path.join(_root, "etc/localtime")
  if os.path.exists(tzpath) and os.path.islink(tzpath):
    log.debug(f"{tzpath} found")
    etctz = os.path.realpath(tzpath)
    start = etctz.find("/") + 1
    while start != 0:
      etctz = etctz[start:]
      try:
        zoneinfo.ZoneInfo(etctz)
        tzinfo = f"{tzpath} is a symlink to"
        found_configs[tzinfo] = etctz.replace(" ", "_")
        # Only need first valid relative path in simlink.
        break
      except zoneinfo.ZoneInfoNotFoundError:
        pass
      start = etctz.find("/") + 1

  if len(found_configs) > 0:
    log.debug(f"{len(found_configs)} found:\n {found_configs}")
    # We found some explicit config of some sort!
    if len(found_configs) > 1:
      # Uh-oh, multiple configs. See if they match:
      unique_tzs = set()
      zoneinfopath = os.path.join(_root, "usr", "share", "zoneinfo")
      directory_depth = len(zoneinfopath.split(os.path.sep))

      for tzname in found_configs.values():
        # Look them up in /usr/share/zoneinfo, and find what they
        # really point to:
        path = os.path.realpath(os.path.join(zoneinfopath, *tzname.split("/")))
        real_zone_name = "/".join(path.split(os.path.sep)[directory_depth:])
        unique_tzs.add(real_zone_name)

      if len(unique_tzs) != 1:
        message = "Multiple conflicting time zone configurations found:\n"
        for key, value in found_configs.items():
          message += f"{key}: {value}\n"
        message += "Fix the configuration, or set the time zone in a TZ environment variable.\n"
        raise zoneinfo.ZoneInfoNotFoundError(message)

    # We found exactly one config! Use it.
    return list(found_configs.values())[0]


def _get_localzone(_root="/"):
  """Creates a timezone object from the timezone name.

  If there is no timezone config, it will try to create a file from the
  localtime timezone, and if there isn't one, it will default to UTC.

  The parameter _root makes the function look for files like /etc/localtime
  beneath the _root directory. This is primarily used by the tests.
  In normal usage you call the function without parameters."""

  # First try the ENV setting.
  tzenv = utils._tz_from_env()
  if tzenv:
    return tzenv

  tzname = _get_localzone_name(_root)
  if tzname is None:
    # No explicit setting existed. Use localtime
    log.debug("No explicit setting existed. Use localtime")
    for filename in ("etc/localtime", "usr/local/etc/localtime"):
      tzpath = os.path.join(_root, filename)

      if not os.path.exists(tzpath):
        continue
      with open(tzpath, "rb") as tzfile:
        tz = zoneinfo.ZoneInfo.from_file(tzfile, key="local")
        break
    else:
      warnings.warn("Can not find any timezone configuration, defaulting to UTC.")
      tz = timezone.utc
  else:
    tz = zoneinfo.ZoneInfo(tzname)

  if _root == "/":
    # We are using a file in etc to name the timezone.
    # Verify that the timezone specified there is actually used:
    utils.assert_tz_offset(tz, error=False)
  return tz


def get_localzone_name():
  """Get the computers configured local timezone name, if any."""
  global _cache_tz_name
  if _cache_tz_name is None:
    _cache_tz_name = _get_localzone_name()

  return _cache_tz_name


def get_localzone():
  """Get the computers configured local timezone, if any."""

  global _cache_tz
  if _cache_tz is None:
    _cache_tz = _get_localzone()

  return _cache_tz


def reload_localzone():
  """Reload the cached localzone. You need to call this if the timezone has changed."""
  global _cache_tz_name
  global _cache_tz
  _cache_tz_name = _get_localzone_name()
  _cache_tz = _get_localzone()

  return _cache_tz



import ast
import inspect
import os
import importlib

from ratio1 import BaseDecentrAIObject, _PluginsManagerMixin

from ratio1_core.core_logging.full_logger import Logger, SBLogger
from ratio1_core.main.entrypoint import main


class Utils(BaseDecentrAIObject, _PluginsManagerMixin):
    pass
  
class MethodExtractor(ast.NodeVisitor):
    def __init__(self):
        self.methods = []

    def visit_FunctionDef(self, node):
        if not node.name.startswith('_'):
            docstring = ast.get_docstring(node)
            args = [arg.arg for arg in node.args.args]
            self.methods.append((node.name, args, docstring))
        self.generic_visit(node)

def extract_methods(source_code):
    tree = ast.parse(source_code)
    extractor = MethodExtractor()
    extractor.visit(tree)
    return extractor.methods

def generate_stub_method(name, args, docstring, is_property=False):
    args_str = ', '.join(args)
    
    method = None
    decorator = '  @property\n' if is_property else ''

    if docstring:
        docstring_lines= docstring.split('\n')
        docstring = '\n'.join([f'    {line}' for line in docstring_lines])
        docstring_formatted = f'    """\n{docstring}\n    """' 
        method = f'{decorator}  def {name}({args_str}):\n{docstring_formatted}\n    raise NotImplementedError\n'
    else:
        method = f'{decorator}  def {name}({args_str}):\n    raise NotImplementedError\n'
    return method

def generate_stub_class(methods):
    stub_methods = [generate_stub_method(name, args, docstring, is_property) for name, args, docstring, is_property in methods]
    
    stub_methods = '\n'.join(stub_methods)
    
    stub_class = f'class CustomPluginTemplate:\n{stub_methods}'
    
    return stub_class

def extract_methods_from_class(cls):
    methods = []
    for name, method in inspect.getmembers(cls, predicate=inspect.isfunction):
        if not name.startswith('_'):
            docstring = inspect.getdoc(method)
            args = list(inspect.signature(method).parameters)
            methods.append((name, args, docstring, False))
    for name, prop in inspect.getmembers(cls, predicate=lambda x: isinstance(x, property)):
        if not name.startswith('_'):
            docstring = inspect.getdoc(prop.fget)
            methods.append((name, ['self'], docstring, True))
    return methods

def get_all_methods(cls):
    methods = {}
    for base_cls in inspect.getmro(cls):
        if base_cls is object:
            continue
        methods.update({name: (args, docstring, is_property) for name, args, docstring, is_property in extract_methods_from_class(base_cls)})
    return [(name, args, docstring, is_property) for name, (args, docstring, is_property) in methods.items()]


def create_stub_file_from_class(cls, destination_filename):
    methods = get_all_methods(cls)
    methods.sort(key=lambda x: x[0])
    stub_class = generate_stub_class(methods)
    with open(destination_filename, 'w', encoding='utf-8') as dest_file:
        dest_file.write(stub_class)

def get_class_from_source(source_code):
    tree = ast.parse(source_code)
    for node in tree.body:
        if isinstance(node, ast.ClassDef):
            return node.name
    return None

def create_stub_file(log:Logger, source_filename, destination_filename):
    module_name = os.path.splitext(source_filename)[0].split(os.sep)
    module_name = '.'.join(module_name[:-1])
    module, class_name, _, _ = Utils(log)._get_module_name_and_class(
        locations=[module_name],
        name=os.path.splitext(os.path.basename(source_filename))[0],
        suffix="Plugin"
        )

    cls = getattr(module, class_name)
    create_stub_file_from_class(cls, destination_filename)
    
# Example usage
log = SBLogger()
full_path = os.path.join('ratio1_core', 'business', 'default', 'custom_exec_01.py')
create_stub_file(log, full_path, 'plugin_template.py')


import os
import json

from time import time
from ratio1.utils.config import log_with_color
from ratio1.const import SESSION_CT, COMMANDS, BASE_CT
from ratio1._ver import __VER__ as version

from pandas import DataFrame
from datetime import datetime


def _get_netstats(
  silent=True,
  online_only=False, 
  allowed_only=False, 
  supervisor=None,
  alias_filter=None,
  supervisors_only=False,
  return_session=False,
  eth=False,
  all_info=False,
  wait_for_node=None
):
  t1 = time()
  from ratio1 import Session
  sess = Session(silent=silent)
  found = None
  if wait_for_node:
    sess.P("Waiting for node '{}' to appear...".format(wait_for_node), color='y')
    found = sess.wait_for_node(wait_for_node, timeout=30)
    if not found:
      sess.P("Node '{}' not found.".format(wait_for_node), color='r')

  dct_info = sess.get_network_known_nodes(
    online_only=online_only, allowed_only=allowed_only, supervisor=supervisor,
    supervisors_only=supervisors_only,
    alias_filter=alias_filter,
    min_supervisors=1,
    eth=eth,
    all_info=all_info, 
  )
  df = dct_info[SESSION_CT.NETSTATS_REPORT]
  supervisor = dct_info[SESSION_CT.NETSTATS_REPORTER]
  super_alias = dct_info[SESSION_CT.NETSTATS_REPORTER_ALIAS]
  nr_supers = dct_info[SESSION_CT.NETSTATS_NR_SUPERVISORS]
  _elapsed = dct_info[SESSION_CT.NETSTATS_ELAPSED] # computed on call
  elapsed = time() - t1 # elapsed=_elapsed
  if return_session:
    return df, supervisor, super_alias, nr_supers, elapsed, sess  
  return df, supervisor, super_alias, nr_supers, elapsed



def get_nodes(args):
  """
  This function is used to get the information about the nodes and it will perform the following:
  
  1. Create a Session object.
  2. Wait for the first net mon message via Session and show progress. 
  3. Wait for the second net mon message via Session and show progress.  
  4. Get the active nodes union via Session and display the nodes marking those peered vs non-peered.
  """
  supervisor_addr = args.supervisor
  alias_filter = args.alias
  online = args.online
  online = True # always online, flag deprecated
  wide = args.wide
  if args.verbose:
    log_with_color(f"Getting nodes from supervisor <{supervisor_addr}>...", color='b')

  res = _get_netstats(
    silent=not args.verbose,
    online_only=online or args.peered,
    allowed_only=args.peered,
    supervisor=supervisor_addr,
    alias_filter=alias_filter,
    eth=args.eth,
    all_info=wide,
    return_session=True,
  )
  df, supervisor, super_alias, nr_supers, elapsed, sess = res
  if args.online:
    FILTERED = ['State']
    df = df[[c for c in df.columns if c not in FILTERED]]

  prefix = "Online n" if (online or args.peered) else "N"
  # network = os.environ.get(BASE_CT.dAuth.DAUTH_NET_ENV_KEY, BASE_CT.dAuth.DAUTH_SDK_NET_DEFAULT)
  network = sess.bc_engine.evm_network
  addr = sess.bc_engine.address
  if supervisor == "ERROR":
    log_with_color(f"No supervisors or no comms available in {elapsed:.1f}s. Please check your settings.", color='r')
  else:
    log_with_color(f"Ratio1 client v{version}: {addr} \n", color='b')
    log_with_color(
      "{}odes on '{}' reported by <{}> '{}' in {:.1f}s ({} supervisors seen):".format(
      prefix, network, supervisor, super_alias, elapsed, nr_supers), 
      color='b'
    )
    import pandas as pd
    pd.set_option('display.float_format', '{:.4f}'.format)
    log_with_color(f"{df}\n")    
  return df
  
  
def get_supervisors(args):
  """
  This function is used to get the information about the supervisors.
  """
  if args.verbose:
    log_with_color("Getting supervisors...", color='b')

  res = _get_netstats(
    silent=not args.verbose,
    online_only=True,
    supervisors_only=True,
    return_session=True,
    eth=args.eth,
    all_info=args.wide,    
  )
  df, supervisor, super_alias, nr_supers, elapsed, sess = res
  FILTERED = ['Oracle', 'State']
  df = df[[c for c in df.columns if c not in FILTERED]]

  import pandas as pd
  pd.set_option('display.float_format', '{:.4f}'.format)
  
  if supervisor == "ERROR":
    log_with_color(f"No supervisors or no comms available in {elapsed:.1f}s. Please check your settings.", color='r')
  else:
    log_with_color(
      "Supervisors on '{}' reported by <{}> '{}' in {:.1f}s".format(
      sess.bc_engine.evm_network, supervisor, super_alias, elapsed), 
      color='b'
    )
    log_with_color(f"{df}")
  return


def get_apps(args):
  """
  Shows the apps running on a given node, if the client is allowed on that node.
  Parameters
  ----------
  args : argparse.Namespace
      Arguments passed to the function.

  """
  verbose = args.verbose
  node = args.node
  show_full = args.full
  as_json = args.json
  owner = args.owner
  wide = args.wide

  # 1. Init session
  from ratio1 import Session
  sess = Session(
    silent=not verbose
  )
  
  res = sess.get_nodes_apps(
    node=node, owner=owner, show_full=show_full, 
    as_json=as_json, as_df=not as_json
  )
  if res is not None:
    network = sess.bc_engine.evm_network
    node_alias = sess.get_node_alias(node) if node else None
    if as_json:
      log_with_color(json.dumps(res, indent=2))
    else:
      df_apps = res
      if df_apps.shape[0] == 0:
        log_with_color(
          "No user apps found on node <{}> '{}' of network '{}'".format(
            node, node_alias, network            
          ), 
          color='r'
        )
        return
      # remove Node column
      if node is not None and owner is None:
        df_apps.drop(columns=['Node'], inplace=True)
      
      if node is None and owner is not None:
        df_apps.drop(columns=['Owner'], inplace=True)
      
      if node is not None:
        last_seen = sess.get_last_seen_time(node)
        last_seen_str = datetime.fromtimestamp(last_seen).strftime('%Y-%m-%d %H:%M:%S') if last_seen else None
        is_online = sess.check_node_online(node)    
        node_status = 'Online' if is_online else 'Offline'
      else:
        last_seen_str = "N/A"
        node_status = "N/A"
      #end if node
      if node == None:
        node = "[All available]"
      by_owner = f" by owner <{owner}>" if owner else ""    
      log_with_color(f"Ratio1 client v{version}:\n", color='b')
      log_with_color(
        "Apps on <{}> ({}) [Status: {}| Last seen: {}]{}:".format(
          node, network, node_status, last_seen_str, by_owner
        ), 
        color='b'
      )
      
      if not wide:        
        df_apps = df_apps[['Node Alias', 'Owner Alias', 'App', 'Plugin', 'Id', 'Probe', 'LastError']]
      log_with_color(f"{df_apps}\n")
    #end if as_json
  #end if res is not None
  return

def _send_command_to_node(args, command, ignore_not_found=False):
  from ratio1 import Session

  node = args.node
  silent = not args.verbose   
  ignore_peering = args.ignore_peering

  t1 = time()
  sess = Session(silent=silent)

  peered, node_addr = sess.is_peered(node, return_full_address=True)
  found = False

  if not ignore_peering:
    found = sess.wait_for_node(node_addr, timeout=30)
    sess.P(f"Node {node_addr} is online.")

  # Display peering status.
  if not peered:
    if not ignore_peering:
      log_with_color(f"Node <{node_addr}> is not peered. Exiting...", color='r')
      return

    if found:
      log_with_color(f"Node '{node}' <{node_addr}> is not peered.", color='r')
    else:
      log_with_color(f"Node '{node}' <{node_addr}> may not accept this command.", color='r')

  if found and peered:
    log_with_color(f"Sending '{command}' to node <{node_addr}>", color='b')
  else:
    log_with_color(f"Sending blind '{command}' to node <{node_addr}>", color='b')

  if (found and peered) or ignore_not_found or ignore_peering:
    if command == COMMANDS.RESTART:
      sess._send_command_restart_node(node_addr)
    elif command == COMMANDS.STOP:
      sess._send_command_stop_node(node_addr)
    else:
      log_with_color(f"Command '{command}' not supported.", color='r')
      return

  sess.close()
  log_with_color(f"Command successfully sent.")
  elapsed = time() - t1
  if not silent:
    log_with_color(f"Command '{command}' seinging took {elapsed}s.", color='b')
  return  

def restart_node(args):
  """
  This function is used to restart the node.
  
  Parameters
  ----------
  args : argparse.Namespace
      Arguments passed to the function.
  """
  node = args.node

  log_with_color(f"Attempting to restart node <{node}>", color='b')
  _send_command_to_node(args, COMMANDS.RESTART, ignore_not_found=True)
  return


def shutdown_node(args):
  """
  This function is used to shutdown the node.
  
  Parameters
  ----------
  args : argparse.Namespace
      Arguments passed to the function.
  """
  node = args.node

  log_with_color(f"Attempting to shutdown node <{node}>", color='b')
  _send_command_to_node(args, COMMANDS.STOP, ignore_not_found=True)
  return


def inspect_node(args):
  """
  This function is used to inspect the node info.

  Parameters
  ----------
  args : argparse.Namespace
      Arguments passed to the function.
  """
  node = args.node
  wide = args.wide
  silent = not args.verbose

  if not silent:
    log_with_color(f"Inspecting node <{node}>", color='b')

  from ratio1 import Session
  session = Session(
    silent=silent
  )
  try:
    node_address = session.get_node_address(node)
    last_hb = session.get_last_hb()
    node_last_hb = last_hb.get(node_address)

    if not node_last_hb:
      # No last HB was found for specified Address, ETH Address or alias.
      raise ValueError("Could not find a node by provided address/alias.")

    # Display node stats from HB:
    _display_node_stats_for_hb(node_last_hb, wide)

  except Exception as e:
    log_with_color(f"An error occurred while trying to get node info: {e}", color='r')
  finally:
    session.close()
  return


def _display_node_stats_for_hb(node_hb, wide=False):
  """
  Display main system information from node heartbeat data.
  
  Parameters
  ----------
  node_hb : dict
      Node heartbeat data containing system information
  wide: bool
      If True, display all plugins.
  """
  from ratio1 import HEARTBEAT_DATA
  from ratio1.utils.config import log_with_color
  from ratio1.const.base import BCctbase
  from ratio1 import PAYLOAD_DATA

  # Extract basic node info
  ee_id = node_hb.get(HEARTBEAT_DATA.EE_ID, 'Unknown')
  node_addr = node_hb.get(HEARTBEAT_DATA.EE_ADDR, 'Unknown')
  eth_addr = node_hb.get(BCctbase.ETH_SENDER, 'Unknown')
  ee_version = node_hb.get(PAYLOAD_DATA.EE_VERSION, 'Unknown')
  uptime = node_hb.get(HEARTBEAT_DATA.UPTIME, 0)
  
  # Display node header
  log_with_color(f"\n=== Node Information ===", color='b')
  log_with_color(f"Node alias: {ee_id}", color='g')
  log_with_color(f"Address: {node_addr}", color='g')
  log_with_color(f"ETH Address: {eth_addr}", color='g')
  log_with_color(f"Edge Node Version: {ee_version}", color='g')
  log_with_color(f"Uptime: {uptime:.1f} seconds ({uptime/3600:.1f} hours)", color='g')
  
  # CPU Information
  cpu_info = node_hb.get(HEARTBEAT_DATA.CPU, 'Unknown')
  cpu_cores = node_hb.get(HEARTBEAT_DATA.CPU_NR_CORES, 0)
  cpu_used = node_hb.get(HEARTBEAT_DATA.CPU_USED, 0)
  
  log_with_color(f"\n=== CPU Information ===", color='b')
  log_with_color(f"Processor: {cpu_info}", color='y')
  log_with_color(f"Cores: {cpu_cores}", color='y')
  log_with_color(f"Usage: {cpu_used:.1f}%", color='y')
  
  # Memory Information
  total_memory = node_hb.get(HEARTBEAT_DATA.MACHINE_MEMORY, 0)
  available_memory = node_hb.get(HEARTBEAT_DATA.AVAILABLE_MEMORY, 0)
  process_memory = node_hb.get(HEARTBEAT_DATA.PROCESS_MEMORY, 0)
  is_alert_ram = node_hb.get(HEARTBEAT_DATA.IS_ALERT_RAM, False)
  
  used_memory = total_memory - available_memory if total_memory and available_memory else 0
  memory_usage_pct = (used_memory / total_memory * 100) if total_memory > 0 else 0
  
  log_with_color(f"\n=== Memory Information ===", color='b')
  log_with_color(f"Total Memory: {total_memory:.2f} GB", color='y')
  log_with_color(f"Available Memory: {available_memory:.2f} GB", color='y')
  log_with_color(f"Used Memory: {used_memory:.2f} GB ({memory_usage_pct:.1f}%)", color='y')
  log_with_color(f"Process Memory: {process_memory:.2f} GB", color='y')
  if is_alert_ram:
    log_with_color(f"RAM Alert: True", color='r')
  
  # Disk Information
  total_disk = node_hb.get(HEARTBEAT_DATA.TOTAL_DISK, 0)
  available_disk = node_hb.get(HEARTBEAT_DATA.AVAILABLE_DISK, 0)
  
  used_disk = total_disk - available_disk if total_disk and available_disk else 0
  disk_usage_pct = (used_disk / total_disk * 100) if total_disk > 0 else 0
  
  log_with_color(f"\n=== Disk Information ===", color='b')
  log_with_color(f"Total Disk: {total_disk:.2f} GB", color='y')
  log_with_color(f"Available Disk: {available_disk:.2f} GB", color='y')
  log_with_color(f"Used Disk: {used_disk:.2f} GB ({disk_usage_pct:.1f}%)", color='y')
  
  # Active Plugins
  active_plugins = node_hb.get(HEARTBEAT_DATA.ACTIVE_PLUGINS, [])
  log_with_color(f"\n=== Active Plugins ===", color='b')
  log_with_color(f"Number of Active Plugins: {len(active_plugins)}", color='y')
  
  if active_plugins:
    plugins_to_display = 99999 if wide else 5
    for i, plugin in enumerate(active_plugins[:plugins_to_display]):  # Show first 5 plugins
      stream_id = plugin.get(HEARTBEAT_DATA.ACTIVE_PLUGINS_INFO.STREAM_ID, 'Unknown')
      signature = plugin.get(HEARTBEAT_DATA.ACTIVE_PLUGINS_INFO.SIGNATURE, 'Unknown')
      instance_id = plugin.get(HEARTBEAT_DATA.ACTIVE_PLUGINS_INFO.INSTANCE_ID, 'Unknown')
      total_payload = plugin.get(HEARTBEAT_DATA.ACTIVE_PLUGINS_INFO.TOTAL_PAYLOAD_COUNT, 0)
      log_with_color(f"  {i+1}. {stream_id} {signature} ({instance_id}) - {total_payload} payloads", color='w')
    
    if len(active_plugins) > plugins_to_display:
      log_with_color(f"  ... and {len(active_plugins) - plugins_to_display} more plugins", color='w')
  
  log_with_color(f"\n", color='w')  # Add final newline for spacing

  return


from ratio1 import Session, Payload
import json
import time


class PayloadHandler:
  def __init__(self):
    self.processed = False
    self.data = None
  
  def on_data(
    self,
    session: Session, 
    node_addr : str, 
    pipeline_name : str, 
    plugin_signature : str, 
    plugin_instance : str,  
    data : Payload      
  ):
    
    if plugin_signature == 'NET_MON_01':
      session.P(
        f"Data from {pipeline_name} ({plugin_signature}, {plugin_instance}) on {node_addr}: {data}",
        color='g'
      )
      with open('./xperimental/netmon/netmon_data.json', 'wt') as f:
        json.dump(data.data, f, indent=2)
        self.data = data.data
        self.processed = True
        
    

if __name__ == '__main__':
  handler_engine = PayloadHandler()  
  
  sess = Session(
    silent=False,
    verbosity=3,
    on_payload=handler_engine.on_data,
    evm_network='mainnet',  # Specify the EVM network if needed
  )
  
  
  while not handler_engine.processed:
    time.sleep(1)
    
  sess.P("Data processing complete.", color='g')
  
  sess.close()  
  


import os
import time

from ratio1 import Session, CustomPluginTemplate, PLUGIN_TYPES

def reply(plugin: CustomPluginTemplate, message: str, user: str):
  """
  This function is used to reply to a message. The given parameters are mandatory
  """
  plugin.obj_cache["users"][user] = self.time()
  plugin.int_cache[user] += 1 # int_cache is a default dict that allows persistence in the plugin
  plugin.P(f"Replying to the {plugin.int_cache[user]} msg of '{user}' on message '{message}'")
  result = f"The answer to your {plugin.int_cache[user]} question is in the question itself: {message}"
  return result


if __name__ == "__main__":   
  session = Session() 
  my_node = os.getenv("EE_TARGET_NODE") 
  assert my_node

  session.wait_for_node(my_node) 
  pipeline, _ = session.create_telegram_simple_bot(
    node=my_node,
    name="telegram_bot_echo",
    message_handler=reply,
  )
  
  pipeline.deploy() # we deploy the pipeline
  session.close() # close the session, leave the pipeline on



class EvmNetData:
  MAINNET = 'mainnet'
  TESTNET = 'testnet'
  DEVNET = 'devnet'

  DAUTH_URL_KEY = 'EE_DAUTH_URL'
  DAUTH_ND_ADDR_KEY = 'EE_DAUTH_ND_ADDR'
  DAUTH_RPC_KEY = 'EE_DAUTH_RPC'
  DAUTH_R1_ADDR_KEY = 'EE_DAUTH_R1_ADDR'
  DAUTH_MND_ADDR_KEY = 'EE_DAUTH_MND_ADDR'
  DAUTH_PROXYAPI_ADDR_KEY = 'EE_DAUTH_PROXYAPI_ADDR'
  DAUTH_POAI_MANAGER_ADDR_KEY = 'EE_DAUTH_POAI_MANAGER_ADDR'
  
  DAUTH_CONTROLLER_ADDR_KEY = 'EE_DAUTH_CONTROLLER_ADDR'  
  DAUTH_GET_ORACLES_ABI = 'EE_DAUTH_GET_ORACLES_ABI'
  
  EE_GENESIS_EPOCH_DATE_KEY = 'EE_GENESIS_EPOCH_DATE'
  EE_EPOCH_INTERVALS_KEY = 'EE_EPOCH_INTERVALS'
  EE_EPOCH_INTERVAL_SECONDS_KEY = 'EE_EPOCH_INTERVAL_SECONDS'
  
  EE_SUPERVISOR_MIN_AVAIL_PRC_KEY = 'EE_SUPERVISOR_MIN_AVAIL_PRC'

  EE_ORACLE_API_URL_KEY = 'EE_ORACLE_API_URL'
  EE_DEEPLOY_API_URL_KEY = 'EE_DEEPLOY_API_URL'
  EE_DAPP_API_URL_KEY = 'EE_DAPP_API_URL_KEY'

# endclass EvmNetData


# Below are basic parameters for plugins that do not interact directly with the blockchain.
class EvmNetConstants:
  EE_NET_MON_01_SUPERVISOR_LOG_TIME_KEY = 'EE_NET_MON_01_SUPERVISOR_LOG_TIME'
  NET_CONFIG_MONITOR_SHOW_EACH_KEY = 'NET_CONFIG_MONITOR_SHOW_EACH'
  SEED_NODES_ADDRESSES_KEY = 'SEED_NODES_ADDRESSES' # This is a list of seed nodes for the network, used for initial connections.
  EE_ORACLE_SYNC_USE_R1FS_KEY = 'EE_ORACLE_SYNC_USE_R1FS'  # This is a boolean flag to use R1FS for oracle sync.
# endclass EvmNetConstants


_DAUTH_ABI_IS_NODE_ACTIVE = [{
  "inputs": [
    {
      "internalType": "address",
      "name": "nodeAddress",
      "type": "address"
    }
  ],
  "name": "isNodeActive",
  "outputs": [
    {
      "internalType": "bool",
      "name": "",
      "type": "bool"
    }
  ],
  "stateMutability": "view",
  "type": "function"
}]


_DAUTH_ABI_GET_ORACLES = [{
  "inputs": [],
  "name": "getOracles",
  "outputs": [
    {
      "internalType": "address[]",
      "name": "",
      "type": "address[]"
    }
  ],
  "stateMutability": "view",
  "type": "function"
}]


_DAUTH_ABI_GET_SIGNERS_deprecated = [{
  "inputs": [],
  "name": "getSigners",
  "outputs": [
    {
      "internalType": "address[]",
      "name": "",
      "type": "address[]"
    }
  ],
  "stateMutability": "view",
  "type": "function"
}]


_GET_NODE_INFO_ABI = [
  {
      "inputs": [
        {
          "internalType": "address",
          "name": "node",
          "type": "address"
        }
      ],
      "name": "getNodeLicenseDetails",
      "outputs": [
        {
          "components": [
            {
              "internalType": "enum LicenseType",
              "name": "licenseType",
              "type": "uint8"
            },
            {
              "internalType": "uint256",
              "name": "licenseId",
              "type": "uint256"
            },
            {
              "internalType": "address",
              "name": "owner",
              "type": "address"
            },
            {
              "internalType": "address",
              "name": "nodeAddress",
              "type": "address"
            },
            {
              "internalType": "uint256",
              "name": "totalAssignedAmount",
              "type": "uint256"
            },
            {
              "internalType": "uint256",
              "name": "totalClaimedAmount",
              "type": "uint256"
            },
            {
              "internalType": "uint256",
              "name": "lastClaimEpoch",
              "type": "uint256"
            },
            {
              "internalType": "uint256",
              "name": "assignTimestamp",
              "type": "uint256"
            },
            {
              "internalType": "address",
              "name": "lastClaimOracle",
              "type": "address"
            },
            {
              "internalType": "bool",
              "name": "isBanned",
              "type": "bool"
            },
            {
              "internalType": "uint256",
              "name": "usdcPoaiRewards",
              "type": "uint256"
            },
            {
              "internalType": "uint256",
              "name": "r1PoaiRewards",
              "type": "uint256"
            }
          ],
          "internalType": "struct LicenseDetails",
          "name": "",
          "type": "tuple"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    }
]


_GET_WALLET_NODES = [
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "wallet",
          "type": "address"
        }
      ],
      "name": "getWalletNodes",
      "outputs": [
        {
          "internalType": "address[]",
          "name": "nodes",
          "type": "address[]"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
]

_GET_ADDRESSES_BALANCES = [
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "addresses",
        "type": "address[]"
      }
    ],
    "name": "getAddressesBalances",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "addr",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "ethBalance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "r1Balance",
            "type": "uint256"
          }
        ],
        "internalType": "struct AddressBalances[]",
        "name": "balances",
        "type": "tuple[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]

# A minimal ERC20 ABI for balanceOf, transfer, and decimals functions.
_ERC20_ABI = [
  {
      "constant": True,
      "inputs": [{"name": "_owner", "type": "address"}],
      "name": "balanceOf",
      "outputs": [{"name": "balance", "type": "uint256"}],
      "payable": False,
      "stateMutability": "view",
      "type": "function"
  },
  {
      "constant": False,
      "inputs": [
          {"name": "_to", "type": "address"},
          {"name": "_value", "type": "uint256"}
      ],
      "name": "transfer",
      "outputs": [{"name": "success", "type": "bool"}],
      "payable": False,
      "stateMutability": "nonpayable",
      "type": "function"
  },
  {
      "constant": True,
      "inputs": [],
      "name": "decimals",
      "outputs": [{"name": "", "type": "uint8"}],
      "payable": False,
      "stateMutability": "view",
      "type": "function"
  }
]

_POAI_MANAGER_ABI = [
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "jobId",
        "type": "uint256"
      }
    ],
    "name": "getJobDetails",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "id",
            "type": "uint256"
          },
          {
            "internalType": "bytes32",
            "name": "projectHash",
            "type": "bytes32"
          },
          {
            "internalType": "uint256",
            "name": "requestTimestamp",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "startTimestamp",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "lastNodesChangeTimestamp",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "jobType",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "pricePerEpoch",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "lastExecutionEpoch",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "numberOfNodesRequested",
            "type": "uint256"
          },
          {
            "internalType": "int256",
            "name": "balance",
            "type": "int256"
          },
          {
            "internalType": "uint256",
            "name": "lastAllocatedEpoch",
            "type": "uint256"
          },
          {
            "internalType": "address[]",
            "name": "activeNodes",
            "type": "address[]"
          },
          {
            "internalType": "address",
            "name": "escrowAddress",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "escrowOwner",
            "type": "address"
          }
        ],
        "internalType": "struct JobWithAllDetails",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "jobId",
        "type": "uint256"
      },
      {
        "internalType": "address[]",
        "name": "newActiveNodes",
        "type": "address[]"
      }
    ],
    "name": "submitNodeUpdate",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "allocateRewardsAcrossAllEscrows",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "oracle",
        "type": "address"
      }
    ],
    "name": "getUnvalidatedJobIds",
    "outputs": [
      {
        "internalType": "uint256[]",
        "name": "",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getIsLastEpochAllocated",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]

# Here are all the constants used for blockchain interaction for each network.
EVM_NET_DATA = {
  EvmNetData.MAINNET: {
    EvmNetData.DAUTH_URL_KEY                    : "https://dauth.ratio1.ai/get_auth_data",
    EvmNetData.DAUTH_CONTROLLER_ADDR_KEY        : "0x90dA5FdaA92edDC80FB73114fb7FE7D97f2be017",
    EvmNetData.DAUTH_ND_ADDR_KEY                : "0xE658DF6dA3FB5d4FBa562F1D5934bd0F9c6bd423",
    EvmNetData.DAUTH_R1_ADDR_KEY                : "0x6444C6c2D527D85EA97032da9A7504d6d1448ecF",
    EvmNetData.DAUTH_MND_ADDR_KEY               : "0x0C431e546371C87354714Fcc1a13365391A549E2",
    EvmNetData.DAUTH_PROXYAPI_ADDR_KEY          : "0xa2fDD4c7E93790Ff68a01f01AA789D619F12c6AC",
    EvmNetData.DAUTH_POAI_MANAGER_ADDR_KEY      : "0xa8d7FFCE91a888872A9f5431B4Dd6c0c135055c1",
    EvmNetData.DAUTH_RPC_KEY                    : "https://base-mainnet.public.blastapi.io",
    EvmNetData.EE_GENESIS_EPOCH_DATE_KEY        : "2025-05-23 16:00:00",
    EvmNetData.EE_EPOCH_INTERVALS_KEY           : 24,
    EvmNetData.EE_EPOCH_INTERVAL_SECONDS_KEY    : 3600,
    EvmNetData.EE_SUPERVISOR_MIN_AVAIL_PRC_KEY  : 0.90,
    EvmNetData.EE_ORACLE_API_URL_KEY            : "https://oracle.ratio1.ai",
    EvmNetData.EE_DEEPLOY_API_URL_KEY           : "https://deeploy-api.ratio1.ai",
    EvmNetData.EE_DAPP_API_URL_KEY              : "https://dapp-api.ratio1.ai",
    EvmNetData.DAUTH_GET_ORACLES_ABI            : _DAUTH_ABI_GET_ORACLES,
  },

  EvmNetData.TESTNET: {
    EvmNetData.DAUTH_URL_KEY                    : "https://testnet-dauth.ratio1.ai/get_auth_data",
    EvmNetData.DAUTH_CONTROLLER_ADDR_KEY        : "0x63BEC1B3004154698830C7736107E7d3cfcbde79",
    EvmNetData.DAUTH_ND_ADDR_KEY                : "0x18E86a5829CA1F02226FA123f30d90dCd7cFd0ED",
    EvmNetData.DAUTH_R1_ADDR_KEY                : "0xCC96f389F45Fc08b4fa8e2bC4C7DA9920292ec64",
    EvmNetData.DAUTH_MND_ADDR_KEY               : "0xa8d7FFCE91a888872A9f5431B4Dd6c0c135055c1",
    EvmNetData.DAUTH_PROXYAPI_ADDR_KEY          : "0xd1c7Dca934B37FAA402EB2EC64F6644d6957bE3b",
    EvmNetData.DAUTH_POAI_MANAGER_ADDR_KEY      : "0x68f825aA8fD4Af498c2998F4b165F103080574d4",
    EvmNetData.DAUTH_RPC_KEY                    : "https://base-sepolia.public.blastapi.io",      
    EvmNetData.EE_GENESIS_EPOCH_DATE_KEY        : "2025-05-23 16:00:00",
    EvmNetData.EE_EPOCH_INTERVALS_KEY           : 24,
    EvmNetData.EE_EPOCH_INTERVAL_SECONDS_KEY    : 3600,
    EvmNetData.EE_SUPERVISOR_MIN_AVAIL_PRC_KEY  : 0.6,
    EvmNetData.EE_ORACLE_API_URL_KEY            : "https://testnet-oracle.ratio1.ai",
    EvmNetData.EE_DEEPLOY_API_URL_KEY           : "https://testnet-deeploy-api.ratio1.ai",
    EvmNetData.EE_DAPP_API_URL_KEY              : "https://testnet-dapp-api.ratio1.ai",
    EvmNetData.DAUTH_GET_ORACLES_ABI            : _DAUTH_ABI_GET_ORACLES,
  },

  
  EvmNetData.DEVNET : {
    EvmNetData.DAUTH_URL_KEY                    : "https://devnet-dauth.ratio1.ai/get_auth_data",
    EvmNetData.DAUTH_CONTROLLER_ADDR_KEY        : "0x46fB56B7499925179d81380199E238f7AE75857B",
    EvmNetData.DAUTH_ND_ADDR_KEY                : "0x90025B92240E3070d5CdcB3f44D6411855c55a73",
    EvmNetData.DAUTH_R1_ADDR_KEY                : "0x277CbD0Cf25F4789Bc04035eCd03d811FAf73691",
    EvmNetData.DAUTH_MND_ADDR_KEY               : "0x17B8934dc5833CdBa1eF42D13D65D677C4727748",
    EvmNetData.DAUTH_PROXYAPI_ADDR_KEY          : "0xFcF04c9A67330431Af75a546615E4881BD8bdC78",
    EvmNetData.DAUTH_POAI_MANAGER_ADDR_KEY      : "0xCc7C4e0f4f25b57807F34227Fb446E68c8c36ce5",
    EvmNetData.DAUTH_RPC_KEY                    : "https://base-sepolia.public.blastapi.io",
    EvmNetData.EE_GENESIS_EPOCH_DATE_KEY        : "2025-06-30 07:00:00",
    EvmNetData.EE_EPOCH_INTERVALS_KEY           : 1,
    EvmNetData.EE_EPOCH_INTERVAL_SECONDS_KEY    : 3600,
    EvmNetData.EE_SUPERVISOR_MIN_AVAIL_PRC_KEY  : 0.6,
    EvmNetData.EE_ORACLE_API_URL_KEY            : "https://devnet-oracle.ratio1.ai",
    EvmNetData.EE_DEEPLOY_API_URL_KEY           : "https://devnet-deeploy-api.ratio1.ai",
    EvmNetData.EE_DAPP_API_URL_KEY              : "https://devnet-dapp-api.ratio1.ai",
    EvmNetData.DAUTH_GET_ORACLES_ABI            : _DAUTH_ABI_GET_ORACLES,
  },
}


# Here are all the constants used for anything unrelated to blockchain interaction
# (e.g. logging parameters for certain admin plugins).
EVM_NET_CONSTANTS = {
  EvmNetData.MAINNET: {
    EvmNetConstants.EE_NET_MON_01_SUPERVISOR_LOG_TIME_KEY: None,  # Default is None, meaning no logging.
    EvmNetConstants.NET_CONFIG_MONITOR_SHOW_EACH_KEY: None,  # Default is None, meaning no periodic logging.
    EvmNetConstants.SEED_NODES_ADDRESSES_KEY: [
      "0xai_Aj1FpPQHISEBelp-tQ8cegwk434Dcl6xaHmuhZQT74if",  # r1s-01
      "0xai_AzySbyf7ggk1UOWkujAy6GFFmDy2MID8Jz7gqxZaDhy8",  # r1s-02
      "0xai_Apkb2i2m8zy8h2H4zAhEnZxgV1sLKAPhjD29B1I_I9z7",  # r1s-03
      "0xai_AgNhMIJZkkWdrTjnmrqdOa6hzAXkoXPNV4Zbvbm_piYJ",  # r1s-04
    ],
    EvmNetConstants.EE_ORACLE_SYNC_USE_R1FS_KEY: False,  # Do not use R1FS for oracle sync in mainnet.
  },
  EvmNetData.TESTNET: {
    EvmNetConstants.EE_NET_MON_01_SUPERVISOR_LOG_TIME_KEY: 120,  # Log every 2 minutes.
    EvmNetConstants.NET_CONFIG_MONITOR_SHOW_EACH_KEY: 120,  # Show every 2 minutes.
    EvmNetConstants.SEED_NODES_ADDRESSES_KEY: [
      "0xai_Amfnbt3N-qg2-qGtywZIPQBTVlAnoADVRmSAsdDhlQ-6",  # tr1s-01
      "0xai_A61eKDV1otIH6uWy3zDlbJNJUWayp1jAsirOuYxztf82",  # tr1s-02
    ],
    EvmNetConstants.EE_ORACLE_SYNC_USE_R1FS_KEY: False,  # Do not use R1FS for oracle sync in testnet.
  },
  EvmNetData.DEVNET: {
    EvmNetConstants.EE_NET_MON_01_SUPERVISOR_LOG_TIME_KEY: 60,  # Log every minute.
    EvmNetConstants.NET_CONFIG_MONITOR_SHOW_EACH_KEY: 60,  # Show every minute.
    EvmNetConstants.SEED_NODES_ADDRESSES_KEY: [
      "0xai_AhIQz47-2dpbncDTODXcP7_cByr0_CI9VEB1dCXnbbG7",  # dr1s-01
      "0xai_AgnygSlY8BwnmaCj6mItg36JHlG_Lh3UqqFaTPbuNzy0",  # dr1s-02
    ],
    EvmNetConstants.EE_ORACLE_SYNC_USE_R1FS_KEY: True,  # Use R1FS for oracle sync in devnet.
  },
}

class EVM_ABI_DATA:
  GET_NODE_INFO = _GET_NODE_INFO_ABI
  GET_WALLET_NODES = _GET_WALLET_NODES
  GET_ADDRESSES_BALANCES = _GET_ADDRESSES_BALANCES
  IS_NODE_ACTIVE = _DAUTH_ABI_IS_NODE_ACTIVE
  ERC20_ABI = _ERC20_ABI
  POAI_MANAGER_ABI = _POAI_MANAGER_ABI
  


import os
import json


from ratio1 import Logger
from ratio1.bc import DefaultBlockEngine



if __name__ == '__main__' :
  
  NETWORKS = [
    "mainnet",
    "testnet",
    "devnet",
  ]
  
  l = Logger("ENC")
  eng = DefaultBlockEngine(
    log=l, name="default", 
    user_config=True,verbosity=10,
  )

  for network in NETWORKS:  
    eng.reset_network(network)

    oracles = eng.web3_get_oracles(debug=True)
    l.P("\nOracles for {}:\n {}".format(
        network, json.dumps(oracles, indent=2)
      ), 
      color='b', show=True
    )  
  
    NODE = "0xED5BE902866855caabF1Acb558209FC40E62524A"
    
    info = eng.web3_get_node_info(node_address=NODE)
    l.P(f"Node info for {NODE}:\n{json.dumps(info, indent=2)}", color='m')
    
    WALLET = "0x464579c1Dc584361e63548d2024c2db4463EdE48"
    try:
      nodes = eng.web3_get_wallet_nodes(address=WALLET)
      l.P(f"Nodes for {WALLET} ({len(nodes)}):")
      for n in nodes:
        if n in oracles:
          l.P(f"  {n} (oracle)", color='g')
        else:
          l.P(f"  {n}")
    except Exception as e:
      l.P(f"Error getting nodes for {WALLET}: {e}", color='r')
  
  
  



import os
import json


from ratio1 import Logger, const
from ratio1.bc import DefaultBlockEngine
from ratio1.utils.config import get_user_folder



if __name__ == '__main__' :
  
  NETWORKS = [
    "mainnet",
    "testnet",
    "devnet",
  ]

  l = Logger("ENC")
  eng = DefaultBlockEngine(
    log=l, name="default", 
    user_config=True
  )
  
  for network in NETWORKS:
    eng.reset_network(network)

    oracles = eng.web3_get_oracles(debug=True)
    l.P("\nOracles for {}:\n {}".format(
        network, json.dumps(oracles, indent=2)
      ), 
      color='b', show=True
    )



"""
R1FS - Ratio1 base IPFS utility functions.


NOTE: 
  - Following the bootstrapping of this module, it takes a few minutes for the relay
    to be connected and the IPFS daemon to be fully operational so sometimes, after
    the start of the engine, the first few `get` operations may fail.


Installation:

1. On the dev node or seed node run ifps_keygen and generate `swarm_key_base64.txt` then
   save this key to the environment variable `EE_SWARM_KEY_CONTENT_BASE64` on the seed 
   oracles as well as in a file.
     
2. On seed node create `ipfs_setup`, copy the files from the `ipfs_setup` including the
  key file.
  
3. Run `setup.sh` on the seed node or:

    ```bash
    #!/bin/bash
    wget https://dist.ipfs.tech/kubo/v0.32.1/kubo_v0.32.1_linux-amd64.tar.gz && \
      tar -xvzf kubo_v0.32.1_linux-amd64.tar.gz && \
      cd kubo && \
      bash install.sh
    ipfs init
    ipfs config --json Swarm.EnableRelayHop true

    ./write_key.sh
    ```
  The `write_key.sh` script should contain the following:
  
    ```bash 
    cat swarm_key_base64.txt | base64 -d > /root/.ipfs/swarm.key
    cat /root/.ipfs/swarm.key
    ```
  
4. Continue on the seed node and run either manually (NOT recommended) or via a systemd
   the ifps daemon using `./launch_service.sh` that basically does:
   
    ```bash
    cp ipfs.service /etc/systemd/system/ipfs.service
    sudo systemctl daemon-reload
    sudo systemctl enable ipfs
    sudo systemctl start ipfs
    ./show.sh
    ```
    
Documentation url: https://docs.ipfs.tech/reference/kubo/cli/#ipfs

"""
import subprocess
import json
from datetime import datetime
import base64
import time
import os
import tempfile
import uuid
import requests
import hashlib
import shutil
import gzip
from io import BytesIO
import ssl
from requests.auth import HTTPBasicAuth
import tempfile

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

DEFAULT_SECRET = "ratio1"

from threading import Lock

__VER__ = "0.2.2"

# empirically determined minimum connection age for IPFS relay
DEFAULT_MIN_CONNECTION_AGE = 0 # seconds


class IPFSCt:
  EE_IPFS_RELAY_ENV_KEY = "EE_IPFS_RELAY"
  EE_SWARM_KEY_CONTENT_BASE64_ENV_KEY = "EE_SWARM_KEY_CONTENT_BASE64"
  EE_IPFS_RELAY_API_KEY = "EE_IPFS_RELAY_API"
  EE_IPFS_API_KEY_BASE64_KEY = "EE_IPFS_API_KEY_BASE64"
  EE_IPFS_CERTIFICATE_BASE64_KEY = "EE_IPFS_CERTIFICATE_BASE64"
  R1FS_DOWNLOADS = "ipfs_downloads"
  R1FS_UPLOADS = "ipfs_uploads"
  CACHE_ROOT = "_local_cache"
  TEMP_DOWNLOAD = os.path.join(f"./{CACHE_ROOT}/_output", R1FS_DOWNLOADS)
  TEMP_UPLOAD = os.path.join(f"./{CACHE_ROOT}/_output", R1FS_UPLOADS)
  
  TIMEOUT = 90 # seconds
  REPROVIDER = "1m"


ERROR_TAG = "Unknown"

COLOR_CODES = {
  "g": "\033[92m",
  "r": "\033[91m",
  "b": "\033[94m",
  "y": "\033[93m",
  "m": "\033[95m",
  'd': "\033[90m", # dark gray
  "reset": "\033[0m"
}

def log_info(msg: str, color="reset", **kwargs):
  timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
  color_code = COLOR_CODES.get(color, COLOR_CODES["reset"])
  reset_code = COLOR_CODES["reset"]
  print(f"{color_code}[{timestamp}] {msg}{reset_code}", flush=True)
  return

class SimpleLogger:
  def P(self, *args, **kwargs):
    log_info(*args, **kwargs)
    return
  
def require_ipfs_started(method):
  """
  decorator to ensure the IPFS is started before executing the method.
  
  parameters
  ----------
  method : callable
      the method to be decorated.
  
  returns
  -------
  callable
      the wrapped method that checks the 'started' attribute.
  
  raises
  ------
  RuntimeError
      if the instance's 'started' attribute is False.
  """
  def wrapper(self, *args, **kwargs):
    if not self.ipfs_started:
      msg = f"R1FS ERROR: {method.__name__} FAILED. R1FS.ipfs_started=={self.ipfs_started}"
      raise RuntimeError(msg)
    return method(self, *args, **kwargs)
  return wrapper  



class R1FSEngine:
  
  _lock: Lock = Lock()
  __instances = {}

  def __new__(
    cls, 
    name: str = "default", 
    logger: any = None, 
    downloads_dir: str = None,
    uploads_dir: str = None,
    base64_swarm_key: str = None,
    ipfs_relay_api: str = None,
    ipfs_api_key_username: str = None,
    ipfs_api_key_password: str = None,
    ipfs_certificate_path: str = None,
    ipfs_relay: str = None,   
    debug=False,     
    min_connection_age: int = DEFAULT_MIN_CONNECTION_AGE,
  ):
    with cls._lock:
      if name not in cls.__instances:
        instance = super(R1FSEngine, cls).__new__(cls)
        instance._build(
          name=name, logger=logger,
          downloads_dir=downloads_dir,
          uploads_dir=uploads_dir,
          base64_swarm_key=base64_swarm_key,
          ipfs_relay=ipfs_relay,
          ipfs_relay_api=ipfs_relay_api,
          ipfs_api_key_username=ipfs_api_key_username,
          ipfs_api_key_password=ipfs_api_key_password,
          ipfs_certificate_path=ipfs_certificate_path,
          debug=debug,
          min_connection_age=min_connection_age,
        )
        cls.__instances[name] = instance
      else:
        instance = cls.__instances[name]
    return instance
  
  
  # base
  if True:
      
    def _build(
      self, 
      name: str = "default",
      logger: any = None, 
      downloads_dir: str = None,
      uploads_dir: str = None,
      base64_swarm_key: str = None, 
      ipfs_relay: str = None,
      ipfs_relay_api: str = None,
      ipfs_api_key_username: str = None,
      ipfs_api_key_password: str = None,
      ipfs_certificate_path: str = None,
      min_connection_age: int = DEFAULT_MIN_CONNECTION_AGE,
      debug=False,     
    ):
      """
      Initialize the IPFS wrapper with a given logger function.
      By default, it uses the built-in print function for logging.
      
      
      """
      self.__DEFAULT_SECRET = DEFAULT_SECRET
      self.__name = name
      if logger is None:
        logger = SimpleLogger()

      self.logger = logger

      self.__ipfs_started = False
      self.__ipfs_address = None
      self.__ipfs_id = None
      self.__ipfs_id_result = None
      self.__min_connection_age = min_connection_age
      self.__connected_at = None
      self.__ipfs_agent = None
      self.__uploaded_files = {}
      self.__downloaded_files = {}
      self.__base64_swarm_key = base64_swarm_key
      self.__ipfs_relay = ipfs_relay
      self.__ipfs_relay_api = ipfs_relay_api
      self.__ipfs_api_key_username = ipfs_api_key_username
      self.__ipfs_api_key_password = ipfs_api_key_password
      self.__ipfs_certificate_path = ipfs_certificate_path
      self.__ipfs_home = None
      self.__downloads_dir = downloads_dir
      self.__uploads_dir = uploads_dir    
      self.__debug = debug
      self.__relay_check_cnt = 0
      
      self.startup()
      return
    
    def startup(self):
      
      if self.__downloads_dir is None:
        if hasattr(self.logger, "get_output_folder"):
          output_folder = self.logger.get_output_folder()
          self.Pd("Using output folder as base: {}".format(output_folder))
          self.__downloads_dir = os.path.join(
            output_folder,
            IPFSCt.R1FS_DOWNLOADS
          )
        else:
          self.__downloads_dir = IPFSCt.TEMP_DOWNLOAD
      #end if downloads_dir    
      os.makedirs(self.__downloads_dir, exist_ok=True)    
      
      if self.__uploads_dir is None:
        if hasattr(self.logger, "get_output_folder"):
          self.__uploads_dir = os.path.join(
            self.logger.get_output_folder(),
            IPFSCt.R1FS_UPLOADS
          )
        else:
          self.__uploads_dir = IPFSCt.TEMP_UPLOAD
      os.makedirs(self.__uploads_dir, exist_ok=True)    

      self.maybe_reset_ipfs()

      self.maybe_start_ipfs(
        base64_swarm_key=self.__base64_swarm_key,
        ipfs_relay=self.__ipfs_relay,
        ipfs_relay_api=self.__ipfs_relay_api,
        ipfs_api_key_username=self.__ipfs_api_key_username,
        ipfs_api_key_password=self.__ipfs_api_key_password,
        ipfs_certificate_path=self.__ipfs_certificate_path
      )
      return
      
      
    def P(self, s, *args, **kwargs):
      s = "[R1FS] " + s
      color = kwargs.pop("color", "d")
      kwargs["color"] = color
      self.logger.P(s, *args, **kwargs)
      return
    
    def Pd(self, s, *args, **kwargs):
      if self.__debug:
        s = "[R1FS][DEBUG] " + s
        color = kwargs.pop("color", "d")
        color = "d" if color != 'r' else "r"
        kwargs["color"] = color
        self.logger.P(s, *args, **kwargs)
      return

    def _hash_secret(self, secret: str) -> bytes:
      secret = str(secret) # to be sure that the passed secret is of string type.
      # Convert text to bytes, then hash with SHA-256 => 32-byte key
      return hashlib.sha256(secret.encode("utf-8")).digest()

  # Private, yet visible (public) helpers.
  if True:
    def _set_debug(self):
      """
      Force debug mode on.
      """
      self.__debug = True
      return

    def _set_min_connection_age(self, min_connection_age: int):
      """
      @Deprecated: Don't use this method anymore, Warm up is not used anymore.
      Set the minimum connection age for IPFS to be considered warmed up.
      """
      return

  # Public properties
  if True:
    @property
    def ipfs_id(self):
      return self.__ipfs_id
    
    @property
    def ipfs_address(self):
      return self.__ipfs_address
    
    @property
    def ipfs_relay(self):
      return self.__ipfs_relay
    
    @property
    def ipfs_agent(self):
      return self.__ipfs_agent
    
    @property
    def ipfs_started(self):
      return self.__ipfs_started

    @property
    def ipfs_connected(self):
      return self.ipfs_started and self.__connected_at is not None

    @property
    def download_folder(self):
      return self.__downloads_dir
    
    
    @property
    def peers(self):
      return self.__peers
    
    @property
    def swarm_peers(self):
      return self.peers
    
    @property
    def uploaded_files(self):
      return self.__uploaded_files
    
    @property
    def downloaded_files(self):
      return self.__downloaded_files
        
    @property
    def connected_at(self):
      return self.__connected_at
    
    @property
    def ipfs_home(self):
      """ return the IPFS home directory from the environment variable IPFS_PATH """
      return os.environ.get("IPFS_PATH")
  
  # boilerplate methods
  if True:
    def _get_unique_name(self, prefix="r1fs", suffix=""):
      str_id = str(uuid.uuid4()).replace("-", "")[:8]
      return f"{prefix}_{str_id}{suffix}"
    
    def _get_unique_upload_name(self, prefix="r1fs", suffix=""):
      return os.path.join(self.__uploads_dir, self._get_unique_name(prefix, suffix))
    
    def _get_unique_or_complete_upload_name(self, fn=None, prefix="r1fs", suffix=""):
      if fn is not None and os.path.dirname(fn) == "":
        return os.path.join(self.__uploads_dir, f"{fn}{suffix}")
      return self._get_unique_upload_name(prefix, suffix=suffix)
    

    def _get_swarm_peers(self):
      peer_lines = []
      try:
        out = subprocess.run(
          ["ipfs", "swarm", "peers"],
          capture_output=True, text=True, timeout=5
        )
        if out.returncode == 0:
          peer_lines = out.stdout.strip().split("\n")
          peer_lines = [line.strip() for line in peer_lines if line.strip()]
          self.Pd(f"Swarm peers: {peer_lines}")
          self.__peers = peer_lines
        else:
          self.__peers = []
        # found = len(self.__peers) > 0
      except Exception as e:
        self.P(f"Error getting swarm peers: {e}", color='r')
      return peer_lines


    def _check_and_record_relay_connection(
      self, 
      max_check_age : int = 3600, 
      debug=False
    ) -> bool:
      """
      Checks if we're connected to the relay_peer_id by parsing 'ipfs swarm peers'.
      If connected and we haven't recorded connected_at yet, we set it.
      """
      log_func = self.P if debug else self.Pd
      relay_found = False
      try:
        if self.connected_at is not None:
          # Already connected, lets see if last check was recent enough:
          elapsed_time = time.time() - self.connected_at
          if elapsed_time < max_check_age:
            relay_found = True
            log_func(f"Relay check #{self.__relay_check_cnt}: Already connected to relay peer for {elapsed_time:.1f}s, skipping check.")
            # If we are connected and the last check was recent enough, return True:
          else:
            log_func(f"Relay check #{self.__relay_check_cnt}: Last connection check was {elapsed_time:.1f}s ago, checking again...")
            # If we are connected but the last check was too long ago, check again:
          #end if needs recheck or not
        #end if connected_at is not None
        if not relay_found:         
          self.__relay_check_cnt += 1
          log_func(f"Relay check #{self.__relay_check_cnt}: Checking IPFS relay connection and swarm peers...")
          peer_lines = self._get_swarm_peers()
          if len(peer_lines) > 0:
            log_func(f"Relay check  #{self.__relay_check_cnt}: {len(peer_lines)} swarm peers.")
            self.__peers = peer_lines
            for line in peer_lines:
              # If the line contains the relay peer ID, we consider ourselves connected:
              if self.__ipfs_relay in line:
                # Record the time if not already set
                relay_found = True
                log_func(f"Relay check #{self.__relay_check_cnt}: Relay ok: {line.strip()}")
                break
              #end if
            #end for
            # now reset the connected_at time if we found the relay peer
            if relay_found:
              # TODO: maybe add first & last connected time
              if self.__connected_at is None:
                # If we found the relay peer and connected_at was not set, set it now:
                self.__connected_at = time.time()
              str_connected = self.logger.time_to_str(self.connected_at)
              self.P(f"Relay check #{self.__relay_check_cnt}: Connected to relay peer recorded at {str_connected}.")
            else:              
              # TODO: maybe add first & last connected time
              # self.__connected_at = None # this is already None or the first connection
              log_func("Relay check #{}: FAIL: relay {} not found in swarm peers:\n{}".format(
                self.__relay_check_cnt, self.__ipfs_relay.split('/')[2],
                json.dumps(peer_lines, indent=2)
                ), color='r'\
              )
            #end if relay_found or not
          #end if len(peer_lines) > 0
      except subprocess.TimeoutExpired:
        self.P(f"Relay check #{self.__relay_check_cnt}: Timeout checking swarm peers.", color='r')
        relay_found = False
      except Exception as e:
        self.P(f"Relay check #{self.__relay_check_cnt}: Error checking swarm peers: {e}", color='r')
        relay_found = False
      #end try
      return relay_found


    def __set_reprovider_interval(self):
      # Command to set the Reprovider.Interval to 1 minute
      cmd = ["ipfs", "config", "--json", "Reprovider.Interval", f'"{IPFSCt.REPROVIDER}"']
      result = self.__run_command(cmd)
      return result


    # TODO: Create a function for setting variables below.
    def __disable_auto_tls(self):
      result = self.__run_command(
        ["ipfs", "config", "--json", "AutoTLS.Enabled", "false"]
      )
      return result

    def __set_routing_type_dht(self):
      result = self.__run_command(
        ["ipfs", "config", "--json", "Routing.Type", '"dht"']
      )
      return result

    def __disable_ws_transport(self):
      result = self.__run_command(
        ["ipfs", "config", "--json", "Swarm.Transports.Network.Websocket", "false"]
      )
      return result

    def __bootstrap_add(self, ipfs_relay):
      result = self.__run_command(
        ["ipfs", "bootstrap", "add", ipfs_relay]
      )
      return result

    # PUBLIC COMMAND
    def get_ipfs_id_data(self) -> dict:
      """
      Get the IPFS peer ID via 'ipfs id' (JSON output).
      Returns the ipfs ID object.
      """
      data = {}
      output = self.__run_command(["ipfs", "id"])  # this will raise an exception if the command fails
      try:
        data = json.loads(output)
      except json.JSONDecodeError:
        raise Exception("Failed to parse JSON from 'ipfs id' output.")
      except Exception as e:
        msg = f"Error getting IPFS ID: {e}. `ipfs id`:\n{data}"
        self.P(msg, color='r')
        raise Exception(f"Error getting IPFS ID: {e}") from e
      return data

    # END PUBLIC COMMANS

    def __run_command(
      self, 
      cmd_list: list, 
      raise_on_error=True,
      timeout=IPFSCt.TIMEOUT,
      verbose=False,
      return_errors=False,
      show_logs=True,
    ):
      """
      Run a shell command using subprocess.run with a timeout.
      Logs the command and its result. If verbose is enabled,
      prints command details. Raises an exception on error if raise_on_error is True.
      """
      failed = False
      output = ""
      errors = ""
      cmd_str = " ".join(cmd_list)
      if show_logs:
        self.Pd(f"Running command: {cmd_str}", color='d')
      try:
        result = subprocess.run(
          cmd_list, 
          capture_output=True, 
          text=True, 
          timeout=timeout,
        )
      except subprocess.TimeoutExpired as e:
        failed = True
        if show_logs:
          self.P(f"Command timed out after {timeout} seconds: {cmd_str}", color='r')
        if raise_on_error:
          raise Exception(f"Timeout expired for '{cmd_str}'") from e
      except Exception as e:
        failed = True
        msg = f"Error running command `{cmd_str}`: {e}"
        if show_logs:
          self.P(msg, color='r')
        if raise_on_error:
          raise Exception(msg) from e
      
      if result.returncode != 0:
        errors = result.stderr.strip()
        failed = True
        if show_logs:
          self.P(f"Command error `{cmd_str}`: {errors}", color='r')
        if raise_on_error:
          raise Exception(f"Error while running '{cmd_str}': {result.stderr.strip()}")
      
      if not failed:
        if show_logs:
          if verbose:
            self.Pd(f"Command output: {result.stdout.strip()}")
        output = result.stdout.strip()
      if return_errors:
        return output, errors
      return output
    

    def __get_id(self) -> str:
      """
      Get the IPFS peer ID via 'ipfs id' (JSON output).
      Returns the 'ID' field as a string.
      """
      data = self.get_ipfs_id_data()
      self.__ipfs_id_result = data
      self.__ipfs_id = data.get("ID", ERROR_TAG)
      self.__ipfs_agent = data.get("AgentVersion", ERROR_TAG)
      addrs = data.get("Addresses", [])
      if not addrs:
        self.__ipfs_address = None
      else:
        self.__ipfs_address = addrs[1] if len(addrs) > 1 else addrs[0] if len(addrs) else ERROR_TAG
      return self.__ipfs_id
    


    @require_ipfs_started
    def __pin_add(self, cid: str) -> str:
      """
      Explicitly pin a CID (and fetch its data) so it appears in the local pinset.
      """
      res = self.__run_command(["ipfs", "pin", "add", cid])
      self.Pd(f"{res}")
      return res  

  # Public R1FS API calls
  if True:
 
    def add_json(
      self, 
      data, 
      fn=None, 
      secret: str = None,
      tempfile=False, 
      show_logs=True,
      raise_on_error=False,
    ) -> bool:
      """
      Add a JSON object to IPFS.
      """
      try:
        json_data = json.dumps(data)
        if tempfile:
          if show_logs:
            self.Pd("Using tempfile for JSON")
          with tempfile.NamedTemporaryFile(
            mode='w', suffix='.json', delete=False
          ) as f:
            f.write(json_data)
          fn = f.name
        else:
          fn = self._get_unique_or_complete_upload_name(fn=fn, suffix=".json")
          if show_logs:
            self.Pd(f"Using unique name for JSON: {fn}")
          with open(fn, "w") as f:
            f.write(json_data)
        #end if tempfile
        cid = self.add_file(
          file_path=fn, secret=secret, show_logs=show_logs,
          raise_on_error=raise_on_error
        )
        return cid
      except Exception as e:
        if show_logs:
          self.P(f"Error adding JSON to IPFS: {e}", color='r')
        return None
      
      
    def add_yaml(
      self, 
      data, 
      fn=None, 
      secret: str = None,
      tempfile=False, 
      show_logs=True,
      raise_on_error=False,
    ) -> str:
      """
      Add a YAML object to IPFS.
      """
      try:
        import yaml
        yaml_data = yaml.dump(data)
        if tempfile:
          if show_logs:
            self.Pd("Using tempfile for YAML")
          with tempfile.NamedTemporaryFile(mode='w', suffix='.yaml', delete=False) as f:
            f.write(yaml_data)
          fn = f.name
        else:
          fn = self._get_unique_or_complete_upload_name(fn=fn, suffix=".yaml")
          if show_logs:
            self.Pd(f"Using unique name for YAML: {fn}")
          with open(fn, "w") as f:
            f.write(yaml_data)
        cid = self.add_file(
          file_path=fn, secret=secret, show_logs=show_logs,
          raise_on_error=raise_on_error
        )
        return cid
      except Exception as e:
        if show_logs:
          self.P(f"Error adding YAML to IPFS: {e}", color='r')
        return None
      
      
    def add_pickle(
      self, 
      data, 
      fn=None, 
      secret: str = None,
      tempfile=False, 
      show_logs=True,
      raise_on_error=False,
    ) -> bool:
      """
      Add a Pickle object to IPFS.
      """
      try:
        import pickle
        if tempfile:
          if show_logs:
            self.Pd("Using tempfile for Pickle")
          with tempfile.NamedTemporaryFile(mode='wb', suffix='.pkl', delete=False) as f:
            pickle.dump(data, f)
          fn = f.name
        else:
          fn = self._get_unique_or_complete_upload_name(fn=fn, suffix=".pkl")
          if show_logs:
            self.Pd(f"Using unique name for pkl: {fn}")
          with open(fn, "wb") as f:
            pickle.dump(data, f)
        cid = self.add_file(
          file_path=fn, secret=secret, show_logs=show_logs,
          raise_on_error=raise_on_error
        )
        return cid
      except Exception as e:
        if show_logs:
          self.P(f"Error adding Pickle to IPFS: {e}", color='r')
        return None


    @require_ipfs_started
    def add_file(
      self,
      file_path: str,
      secret: str = None,
      raise_on_error: bool = False,
      show_logs: bool = True
    ) -> str:
      """
      Add a file to R1FS with default encryption. The secret parameter is mandatory,
      defaulting to 'ratio1'. Each encryption run is chunked for large files, and
      the original filename is stored in JSON metadata.


      Parameters
      ----------
      file_path : str
        Path to the local plaintext file.

      secret : str, optional
        Mandatory passphrase, defaulting to 'ratio1'. Must not be empty.
        
      raise_on_error : bool, optional
        If True, raise an Exception on command errors. Otherwise logs them. Default is False.

      show_logs : bool, optional
        Whether to show logs via self.P / self.Pd. Default is True.

      Returns
      -------
      str
        The folder CID of the wrapped IPFS directory containing the ciphertext.

      Raises
      ------
      FileNotFoundError
        If file_path does not exist.

      ValueError
        If secret is empty.

      RuntimeError
        If the 'ipfs add' command yields no output.

      Examples
      --------
      >>> cid = engine.add_file("/data/large_model.bin")
      >>> print(cid)
      QmFolder123ABC
      """
      if secret in ["", None]:
        secret = self.__DEFAULT_SECRET
      
      if not os.path.isfile(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")

      # Check file size and throw an error if larger than 2 GB.
      file_size = os.path.getsize(file_path)
      if file_size > 2 * 1024 * 1024 * 1024:
        raise ValueError(f"File {file_path} is too large ({file_size} bytes). Maximum allowed size is 2 GB.")

      key = self._hash_secret(secret)  # mandatory passphrase
      nonce = os.urandom(12)           # recommended for GCM
      original_basename = os.path.basename(file_path)

      # JSON metadata storing the original filename
      meta_dict = {"filename": original_basename}
      meta_bytes = json.dumps(meta_dict).encode("utf-8")

      tmp_cipher_path = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex + ".bin")
      
      folder_cid = None
      start_time = time.time()
      try:
        encryptor = Cipher(algorithms.AES(key), modes.GCM(nonce)).encryptor()

        chunk_size = 1024 * 1024  # 1 MB chunks
        with open(file_path, "rb") as fin, open(tmp_cipher_path, "wb") as fout:
          # [nonce][4-byte-len][metadata][ciphertext][16-byte GCM tag]
          fout.write(nonce)
          meta_len = len(meta_bytes)
          fout.write(meta_len.to_bytes(4, "big"))
          # encrypt the metadata and save
          enc_meta_data = encryptor.update(meta_bytes)
          fout.write(enc_meta_data)

          while True:
            chunk = fin.read(chunk_size)
            if not chunk:
              break
            fout.write(encryptor.update(chunk))
          #end while there are still bytes to read
          final_ct = encryptor.finalize()
          fout.write(final_ct)
          # Append 16-byte GCM tag
          tag = encryptor.tag
          fout.write(tag)
        #end with fin, fout
        
        # Now we IPFS-add the ciphertext
        output = self.__run_command(["ipfs", "add", "-q", "-w", tmp_cipher_path], show_logs=show_logs)
        lines = output.strip().split("\n")
        if not lines:
          raise RuntimeError("No output from 'ipfs add -w -q' for ciphertext.")
        folder_cid = lines[-1].strip()
      except Exception as e:
        msg = f"Error encrypting file {file_path}: {e}"
        if raise_on_error:
          raise RuntimeError(msg)
        else:
          self.P(msg, color='r')
      #end try
      elapsed_time = time.time() - start_time
      # Cleanup temp file
      os.remove(tmp_cipher_path)
      
      if folder_cid is not None:
        if self.__ipfs_relay_api is not None:
          #  Notifying the Relay about a new CID.
          try:
            request_url = f"{self.__ipfs_relay_api}/api/v0/pin/add?arg={folder_cid}"
            response = requests.post(request_url,
                                     auth=HTTPBasicAuth(self.__ipfs_api_key_username, self.__ipfs_api_key_password),
                                     verify=self.__ipfs_certificate_path)
            if response.status_code == 200:
              self.Pd(f"Relay successfully notified about CID={folder_cid}.")
            else:
              msg = f"Failed to notify relay about CID {folder_cid}: {response.text}"
              if raise_on_error:
                raise RuntimeError(msg)
              else:
                self.P(msg, color='r')
            #end if response status code
          except requests.RequestException as e:
            msg = f"Error notifying relay about CID {folder_cid}: {e}"
            if raise_on_error:
              raise RuntimeError(msg)
            else:
              self.P(msg, color='r')
          #end try

        self.__uploaded_files[folder_cid] = file_path
        # now we pin the folder
        res = self.__pin_add(folder_cid)
        if show_logs:
          self.P(f"Added file {file_path} as <{folder_cid}> in {elapsed_time:.1f}s")      
        #end if show_logs
      #end if folder_cid is not None
      return folder_cid


    @require_ipfs_started
    def get_file(
      self,
      cid: str,
      local_folder: str = None,
      secret: str = None,
      timeout: int = None,
      pin: bool = True,
      raise_on_error: bool = False,
      show_logs: bool = True,
      return_absolute_path: bool = True,
    ) -> str:
      """
      Retrieve an encrypted file from R1FS by CID, decrypt with AES-GCM in streaming mode.
      The secret parameter is mandatory (default 'ratio1'). The original filename is
      restored from JSON metadata.

      R1FS get can time out if it stalls. If the timeout is None, we use IPFSCt.TIMEOUT.

      Parameters
      ----------
      cid : str
        The folder CID (wrapped single file).

      local_folder : str, optional
        Destination folder. If None, we default to something like self.__downloads_dir/<CID>.

      secret : str, optional
        Passphrase for AES-GCM. Must not be empty. Defaults to 'ratio1'.

      timeout : int, optional
        Maximum seconds for the IPFS get. If None, use IPFSCt.TIMEOUT.

      pin : bool, optional
        If True, we optionally pin the folder. Default True.

      raise_on_error : bool, optional
        If True, raise an Exception on command errors/timeouts. Otherwise logs them. Default False.

      show_logs : bool, optional
        If True, logs steps via self.P / self.Pd. Default True.

      return_absolute_path : bool, optional
        If True, return the absolute path to the restored plaintext file.

      Returns
      -------
      str
        The full path to the restored plaintext file.

      Raises
      ------
      ValueError
        If the secret is empty.

      RuntimeError
        If multiple or zero files are found in the downloaded folder,
        or if we fail to parse the JSON metadata.

      Exception
        If the GCM tag is invalid or the IPFS command times out
        and raise_on_error=True.

      Examples
      --------
      >>> # Simple usage with default passphrase
      >>> local_file = engine.get_file("QmEncFolderXYZ")
      >>> print(local_file)
      /app/downloads/QmEncFolderXYZ/original_filename.bin
      
      """
      # Validate CID parameter
      if cid in [None, ""]:
        msg = "CID parameter cannot be None or empty"
        if raise_on_error:
          raise ValueError(msg)
        else:
          if show_logs:
            self.P(msg, color='r')
          return None
      
      if secret in ["", None]:
        secret = self.__DEFAULT_SECRET
        
      key = self._hash_secret(secret)

      if pin:
        try:
          pin_result = self.__pin_add(cid)
        except Exception as e:
          msg = f"Error pinning CID {cid}: {e}"
          if raise_on_error:
            raise RuntimeError(msg)
          else:
            self.P(msg, color='r')
            return None
          # end if
        #end try
      #end if pin

      if local_folder is None:
        local_folder = self.__downloads_dir # default downloads directory
      
      os.makedirs(local_folder, exist_ok=True)
      
      local_folder = os.path.join(local_folder, cid) # add the CID as a subfolder
      
      # if the folder exists cleanup the content 
      if os.path.exists(local_folder):
        if show_logs:
          files = os.listdir(local_folder)
          self.Pd(f"Cleaning up {local_folder} with {files}")
        #end if show_logs
        shutil.rmtree(local_folder)
      #end if local_folder exists

      
      if show_logs:
        self.Pd(f"Downloading file {cid} to {local_folder}")
      # IPFS get the single ciphertext file
      ipfs_timeout = timeout if timeout else 90  # or IPFSCt.TIMEOUT
      start_time = time.time()
      self.__run_command(
        ["ipfs", "get", cid, "-o", local_folder],
        timeout=ipfs_timeout,
        raise_on_error=raise_on_error,
        show_logs=show_logs
      )
      download_elapsed_time = time.time() - start_time

      # Expect exactly one file
      contents = os.listdir(local_folder)
      if len(contents) != 1:
        msg = f"Expected 1 file in {local_folder}, found {contents}"
        if raise_on_error:
          raise RuntimeError(msg)
        else:
          self.P(msg, color='r')
          return

      cipher_path = os.path.join(local_folder, contents[0])
      
      out_path = None

      # Decrypt with AES-GCM
      start_time = time.time()
      try:
        with open(cipher_path, "rb") as fin:
          nonce = fin.read(12)
          meta_len_bytes = fin.read(4)
          meta_len = int.from_bytes(meta_len_bytes, "big")

          decryptor = Cipher(algorithms.AES(key), modes.GCM(nonce)).decryptor()
          
          # Read the metadata and decrypt it
          enc_meta_data = fin.read(meta_len)
          meta_data = decryptor.update(enc_meta_data) # TODO: verify if this is correct
          meta_dict = json.loads(meta_data.decode("utf-8"))

          original_filename = meta_dict.get("filename", "restored_file.bin")

          # File size + chunk logic to isolate last 16 bytes as GCM tag
          fin.seek(0, os.SEEK_END)
          total_size = fin.tell()
          data_start = 12 + 4 + meta_len
          tag_size = 16
          content_size = total_size - data_start - tag_size
          fin.seek(data_start, os.SEEK_SET)

          out_path = os.path.join(local_folder, original_filename)
          if return_absolute_path:
            out_path = os.path.abspath(out_path)
          chunk_size = 1024 * 1024

          with open(out_path, "wb") as fout:
            remaining = content_size
            while remaining > 0:
              read_len = min(chunk_size, remaining)
              chunk = fin.read(read_len)
              if not chunk:
                break
              fout.write(decryptor.update(chunk))
              remaining -= read_len
            #end while there are still bytes to read
            # Final 16 bytes => GCM tag
            tag = fin.read(16)
            # decryptor.authenticate_tag(tag)
            # final_pt = decryptor.finalize()
            final_pt = decryptor.finalize_with_tag(tag)
            if final_pt:
              fout.write(final_pt)
          #end with fout 
        #end with fin
        decrypt_elapsed_time = time.time() - start_time
      except Exception as e:
        out_path = None
        msg = f"Error decrypting file {cipher_path}: {e}"
        if raise_on_error:
          raise RuntimeError(msg)
        else:
          self.P(msg, color='r')
      #end try

      # Optionally remove the ciphertext
      os.remove(cipher_path)      

      if out_path:
        if show_logs:
          self.P(f"Downloaded/descrypted in {download_elapsed_time:.1f}s/{decrypt_elapsed_time:.1f}s <{cid}> to {out_path}")
        self.__downloaded_files[cid] = out_path      
      #end if out_path is not None
      return out_path


    @require_ipfs_started
    def list_pins(self):
      """
      List pinned CIDs via 'ipfs pin ls --type=recursive'.
      Returns a list of pinned CIDs.
      """
      output = self.__run_command(["ipfs", "pin", "ls", "--type=recursive"])
      pinned_cids = []
      for line in output.split("\n"):
        line = line.strip()
        if not line:
          continue
        parts = line.split()
        if len(parts) > 0:
          pinned_cids.append(parts[0])
      return pinned_cids
    
    
    @require_ipfs_started
    def is_cid_available(self, cid: str, max_wait=3) -> bool:
      """
      Check if a CID is available on IPFS.
      Returns True if the CID is available, False otherwise.
      
      Parameters
      ----------
      cid : str
          The CID to check.
          
      max_wait : int
          The maximum time to wait for the CID to be found.
          
      """
      CMD = ["ipfs", "block", "stat", cid]  
      result = True
      try:
        res = self.__run_command(CMD, timeout=max_wait)
        self.Pd(f"{cid} is available:\n{res}")
      except Exception as e:
        result = False
      return result
      

  # Start/stop IPFS methods (R1FS API)
  if True:
    @property
    def is_ipfs_warmed(self) -> bool:
      """
      @Deprecated:
      This method is deprecated and will be removed in future versions.
      Check if IPFS is warmed up (connected to the relay and has been for a while).
      """
      return True

    def is_ipfs_daemon_running(
      self,
      host="127.0.0.1",
      port=5001,
      method="POST",
      timeout=3
    ) -> bool:
      """
      Checks if an IPFS daemon is running by calling /api/v0/version
      on the specified host and port. Some configurations require
      POST instead of GET, so we allow a method argument.

      Returns:
          bool: True if the IPFS daemon responds successfully; False otherwise.
      """
      url = f"http://{host}:{port}/api/v0/version"
      result = False
      output = None
      try:
        if method.upper() == "POST":
          response = requests.post(url, timeout=timeout)
        else:
          response = requests.get(url, timeout=timeout)

        if response.status_code == 200:
          data = response.json()
          output = str(data)
          if "Version" in data:
            result = True
          else:
            result = False
        else:
          result = False
          output = f"Status code: {response.status_code}"

      except Exception as e:
        result = False
        output = str(e)
      self.P(f"IPFS daemon run-check: {result} ({output})")
      return result        
    
    def is_ipfs_daemon_ready(self, max_wait=30, step=1):
      """ Check with timeout if the IPFS daemon is running and ready to accept requests."""
      waited = 0
      while waited < max_wait:
        if self.is_ipfs_daemon_running():
          return True
        time.sleep(step)
        waited += step
      return False    
    
    def maybe_reset_ipfs(self):
      """ Reset the IPFS repository if needed, remove swarm key and ipfs home."""
      

    def maybe_start_ipfs(
      self, 
      base64_swarm_key: str = None, 
      ipfs_relay: str = None,
      ipfs_relay_api: str = None,
      ipfs_api_key_username: str = None,
      ipfs_api_key_password: str = None,
      ipfs_certificate_path: str = None,
    ) -> bool:
      """
      This method initializes the IPFS repository if needed, connects to a relay, and starts the daemon.
      TODO: (Vitalii) Split this into smaller methods.
      """
      if self.ipfs_started:
        return
      
      self.P("Starting R1FS...", color='m')
      
      if base64_swarm_key is None:
        base64_swarm_key = os.getenv(IPFSCt.EE_SWARM_KEY_CONTENT_BASE64_ENV_KEY)
        if base64_swarm_key is not None:
          self.P(f"Found env IPFS swarm key: {str(base64_swarm_key)[:4]}...", color='d')
          if len(base64_swarm_key) < 10:
            self.P(f"Invalid IPFS swarm key: `{base64_swarm_key}`", color='r')
            return False
        
      if ipfs_relay is None:
        ipfs_relay = os.getenv(IPFSCt.EE_IPFS_RELAY_ENV_KEY)
        if ipfs_relay is not None:
          self.P(f"Found env IPFS relay: {ipfs_relay}", color='d')
          if len(ipfs_relay) < 10:
            self.P(f"Invalid IPFS relay: `{ipfs_relay}`", color='r')
            return False

      if ipfs_relay_api is None:
        ipfs_relay_api = os.getenv(IPFSCt.EE_IPFS_RELAY_API_KEY)
        if ipfs_relay_api is not None:
          self.P(f"Found env IPFS relay API: {ipfs_relay_api}", color='d')

      # Set up IPFS API key username and password.
      if ipfs_api_key_username is None or ipfs_api_key_password is None:
        try:
          ipfs_api_key_base64 = os.getenv(IPFSCt.EE_IPFS_API_KEY_BASE64_KEY)
          ipfs_api_key_b = base64.b64decode(ipfs_api_key_base64)
          ipfs_api_key = str(ipfs_api_key_b, 'utf-8')
          split_api_key = ipfs_api_key.split(":")
          ipfs_api_key_username = split_api_key[0]
          ipfs_api_key_password = split_api_key[1]
        except Exception as e:
          self.P(f"An error occurred while extracting IPFS Relay username and password {e}", color='r')


      # Set up certificate
      if ipfs_certificate_path is None:
        try:
          certificate_encoded = os.getenv(IPFSCt.EE_IPFS_CERTIFICATE_BASE64_KEY)
          decoded = self.__decode_base64_gzip_to_text(certificate_encoded)

          with tempfile.NamedTemporaryFile(delete=False, mode='w', suffix='.crt') as f:
            f.write(decoded)
            ipfs_certificate_path = f.name

          if not os.path.isfile(ipfs_certificate_path):
            raise FileNotFoundError(f"Could not create certificate file : {ipfs_certificate_path}")
        except Exception as e:
          self.P(f"Error reading the certificate file: {e}", color='r')

      if not base64_swarm_key or not ipfs_relay:
        self.P("Missing env values EE_SWARM_KEY_CONTENT_BASE64 and EE_IPFS_RELAY.", color='r')
        return False
      
      self.__base64_swarm_key = base64_swarm_key
      self.__ipfs_relay = ipfs_relay
      self.__ipfs_relay_api = ipfs_relay_api
      self.__ipfs_api_key_username = ipfs_api_key_username
      self.__ipfs_api_key_password = ipfs_api_key_password
      self.__ipfs_certificate_path = ipfs_certificate_path
      hidden_base64_swarm_key = base64_swarm_key[:8] + "..." + base64_swarm_key[-8:]
      
      existing_ipfs_home = os.getenv("IPFS_PATH")
      home_ready = False
      if existing_ipfs_home:
        self.P(f"Found existing IPFS home: {existing_ipfs_home}", color='d')
        if IPFSCt.CACHE_ROOT in existing_ipfs_home and os.path.isdir(existing_ipfs_home):
          self.__ipfs_home = os.path.abspath(existing_ipfs_home)
          home_ready = True
        else:
          self.P(f"Invalid IPFS home: {existing_ipfs_home}", color='r')
        #endif
      #endif 
      if not home_ready:
        self.P("No existing IPFS home found, creating a new one.", color='d')
        ipfs_home = os.path.join(self.logger.base_folder, ".ipfs/")        
        os.makedirs(ipfs_home, exist_ok=True)
        self.__ipfs_home = os.path.abspath(ipfs_home)
        os.environ["IPFS_PATH"] = self.__ipfs_home
      
      config_path = os.path.join(self.__ipfs_home, "config")
      swarm_key_path = os.path.join(self.__ipfs_home, "swarm.key")

      msg = f"Starting R1FS <{self.__name}>:"
      msg += f"\n  IPFS Home: {self.__ipfs_home}"
      msg += f"\n  Relay:    {self.__ipfs_relay}"
      msg += f"\n  Download: {self.__downloads_dir}"
      msg += f"\n  Upload:   {self.__uploads_dir}"
      msg += f"\n  SwarmKey: {hidden_base64_swarm_key}"
      msg += f"\n  Debug:    {self.__debug}"
      msg += f"\n  Repo:     {self.ipfs_home}"
      self.P(msg, color='d')
      
      # Write the swarm key at every start.
      try:
        decoded_key = base64.b64decode(base64_swarm_key)
        with open(swarm_key_path, "wb") as f:
          f.write(decoded_key)
        os.chmod(swarm_key_path, 0o600)
        self.P("Swarm key written successfully.", color='g')
      except Exception as e:
        self.P(f"Error writing swarm.key: {e}", color='r')
        return False

      if not os.path.isfile(config_path):
        # Repository is not initialized; init.
        try:
          self.P("Initializing IPFS repository...")
          self.__run_command(["ipfs", "init"])
        except Exception as e:
          self.P(f"Error during IPFS init: {e}", color='r')
          return False
      else:
        self.P(f"IPFS repository already initialized in {config_path}.", color='g')


      # Check if daemon is already running by attempting to get the node id.
      try:
        self.P("Trying to see if IPFS daemon is running...", color='d')
        n_ipfs_daemon_checks = 3
        for attempt in range(1, n_ipfs_daemon_checks + 1):
          ipfs_daemon_running = self.is_ipfs_daemon_running()
          if ipfs_daemon_running:
            break
          else:
            self.P(f"Check {attempt}/{n_ipfs_daemon_checks} IPFS started: {ipfs_daemon_running}", color='d')
            time.sleep(2)
        #end for
        
        if ipfs_daemon_running:
          self.P("IPFS daemon already running", color='g')
        else:
          # If not running, start the daemon in the background.
          self.P("IPFS daemon not running. Trying to start...", color='r')     
          ###################################################
          #######             CLEANUP PHASE            ######
          ###################################################                            
          # we start by removing any existing bootstrap nodes
          try:
            self.P("Removing public IPFS bootstrap nodes...")
            self.__run_command(["ipfs", "bootstrap", "rm", "--all"])
          except Exception as e:
            self.P(f"Error removing bootstrap nodes: {e}", color='r')
               
          # then delete the repository lock file if it exists
          lock_file = os.path.join(self.__ipfs_home, "repo.lock")
          if os.path.isfile(lock_file):
            self.P(f"Deleting lock file {lock_file}...")
            os.remove(lock_file)
          else:
            self.P(f"Lock file {lock_file} not found.")
          # next check if the api is zero-length - if so, delete it
          api_file = os.path.join(self.__ipfs_home, "api")
          if os.path.isfile(api_file) and os.path.getsize(api_file) == 0:
            self.P(f"Deleting zero-length api file {api_file}...")
            os.remove(api_file)
          #endif            
          # now we can start the daemon
          ###################################################
          #######        END OF CLEANUP PHASE        ########
          ###################################################
          self.__set_reprovider_interval()
          self.__disable_auto_tls()
          self.__set_routing_type_dht()
          self.__disable_ws_transport()
          self.__bootstrap_add(self.__ipfs_relay)
          self.P("Starting IPFS daemon in background...")
          subprocess.Popen(["ipfs", "daemon", "--enable-gc", "--migrate=true"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
          max_attempts = 10
          sleep_time = 2
          for attempt in range(max_attempts):
            ipfs_daemon_running = self.is_ipfs_daemon_running()
            if ipfs_daemon_running:
              self.P("IPFS daemon started successfully.", color='g')
              break
            else:
              self.P(f"Check {attempt + 1}/{max_attempts} IPFS running: {ipfs_daemon_running}", color='r')
              time.sleep(sleep_time)
          #end for
          if not ipfs_daemon_running:
            self.P("Failed to start IPFS daemon after multiple attempts.", color='r')
            return False
        #end if daemon running
      except Exception as e:
        self.P(f"Error starting IPFS daemon: {e}", color='r')
        return

      # last phase: connect to the relay      
      try:
        relay_ip = ipfs_relay.split("/")[2]
        self.P("Getting the IPFS ID...")
        my_id = self.__get_id()
        assert my_id != ERROR_TAG, "Failed to get IPFS ID."
        self.P("Checking swarm peers...")
        swarm_peers = self._get_swarm_peers()
        if len(swarm_peers) > 0:
          self.P(f"{len(swarm_peers)} swarm peers detected. Checking for relay connection...")
          relay_found = self._check_and_record_relay_connection(debug=True)
          if relay_found:
            self.__ipfs_started = True
            self.P(f"{my_id} connected to: {relay_ip}", color='g', boxed=True)
          else:
            self.P("No relay connection found in swarm peers.", color='r')
            self.__ipfs_started = False
          #end if relay_found
        
        if not self.__ipfs_started:
          msg =  f"Connecting to R1FS relay"
          msg += f"\n  IPFS Home:  {self.ipfs_home}"
          msg += f"\n  IPFS ID:    {my_id}"
          msg += f"\n  IPFS Addr:  {self.__ipfs_address}"
          msg += f"\n  IPFS Agent: {self.__ipfs_agent}"
          msg += f"\n  Relay:      {ipfs_relay}"
          self.P(msg, color='m')
          result = self.__run_command(["ipfs", "swarm", "connect", ipfs_relay])
          if "connect" in result.lower() and "success" in result.lower():
            self.P(f"{my_id} connected to: {relay_ip}", color='g', boxed=True)
            self.__ipfs_started = True
            self.P("Re-checking swarm peers...")
            swarm_peers = self._get_swarm_peers()
            self.P(f"Swarm peers:\n {json.dumps(swarm_peers, indent=2)}")
            self._check_and_record_relay_connection(debug=True)
          else:
            self.P("Relay connection result did not indicate success.", color='r')
      except Exception as e:
        self.P(f"Error connecting to relay: {e}", color='r')
      #end try
      return self.ipfs_started

  def __decode_base64_gzip_to_text(self, encoded_str):
    try:
      # Step 1: Decode base64
      compressed_data = base64.b64decode(encoded_str)
      # Step 2: Decompress gzip
      with gzip.GzipFile(fileobj=BytesIO(compressed_data)) as f:
        decompressed_data = f.read()
      # Step 3: Convert bytes to string
      return decompressed_data.decode('utf-8')
    except Exception as e:
      self.P(f"Error decoding base64 string: {e}")
    return ''

if __name__ == '__main__':
  from ratio1 import Logger
  log = Logger("IPFST")
  # eng = R1FSEngine()


import os
import io
import numpy as np
import base64

from time import time as tm
from shutil import copyfile
from io import BytesIO

try:
  import cv2
except:
  pass

class _ComputerVisionMixin(object):
  """
  Mixin for computer vision functionalities that are attached to `libraries.logger.Logger`

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_ComputerVisionMixin, self).__init__()
    return

  @staticmethod
  def is_image(file):
    return any([file.endswith(tail) for tail in ('.png', '.jpg', '.jpeg')])

  @staticmethod
  def increase_brightness(img, value=30):
    hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h, s, v = cv2.split(hsv)

    lim = 255 - value
    v[v > lim] = 255
    v[v <= lim] += value

    final_hsv = cv2.merge((h, s, v))
    img = cv2.cvtColor(final_hsv, cv2.COLOR_HSV2BGR)
    return img

  @staticmethod
  def convert_gray(np_src, copy=False):
    np_source = np_src.copy() if copy else np_src
    if len(np_source.shape) == 3 and np_source.shape[-1] != 1:
      np_source = cv2.cvtColor(np_source, cv2.COLOR_BGR2GRAY)
    return np_source

  @staticmethod
  def scale_image(np_src, copy=False):
    np_source = np_src.copy() if copy else np_src
    if np_source.max() > 1:
      np_source = (np_source / 255).astype(np.float32)
    # endif
    return np_source

  @staticmethod
  def center_image(image, target_h, target_w, copy=False):
    image = image.copy() if copy else image
    new_h, new_w, _ = image.shape

    # determine the new size of the image
    if (float(target_w) / new_w) < (float(target_h) / new_h):
      new_h = int((new_h * target_w) / new_w)
      new_w = int(target_w)
    else:
      new_w = int((new_w * target_h) / new_h)
      new_h = int(target_h)

    # embed the image into the standard letter box
    new_image = np.ones((target_h, target_w, 3), dtype=np.uint8)
    top = int((target_h - new_h) // 2)
    bottom = int((target_h + new_h) // 2)
    left = int((target_w - new_w) // 2)
    right = int((target_w + new_w) // 2)

    resized = cv2.resize(image, dsize=(new_w, new_h))
    new_image[top:bottom, left:right, :] = resized
    return new_image

  @staticmethod
  def center_image_coordinates(src_h, src_w, target_h, target_w):
    asp_src = src_h / src_w
    asp_dst = target_h / target_w
    if asp_src > asp_dst:
      # if src_h > src_w:
      new_h = target_h
      new_w = int(target_h / (src_h / src_w))
    else:
      new_w = target_w
      new_h = int(target_w / (src_w / src_h))
    # endif

    left = target_w // 2 - new_w // 2
    top = target_h // 2 - new_h // 2
    right = left + new_w
    bottom = top + new_h
    return (top, left, bottom, right), (new_h, new_w)

  @staticmethod
  def center_image2(np_src, target_h, target_w, copy=False):
    np_source = np_src.copy() if copy else np_src
    shape = (target_h, target_w, np_src.shape[-1]) if len(np_src.shape) == 3 else (target_h, target_w)
    (top, left, bottom, right), (new_h, new_w) = _ComputerVisionMixin.center_image_coordinates(
      np_source.shape[0], np_source.shape[1],
      target_h, target_w
    )
    np_dest = np.zeros(shape).astype(np.float32 if np_src.max() <= 1 else np.uint8)
    np_src_mod = cv2.resize(np_src, dsize=(new_w, new_h))
    np_dest[top:bottom, left:right] = np_src_mod
    return np_dest

  @staticmethod
  def rescale_boxes(boxes,
                    src_h,
                    src_w,
                    target_h,
                    target_w,
                    ):
    if np.max(boxes) <= 1:
      # [0:1] to [0:model_shape]
      boxes[:, 0] *= target_h
      boxes[:, 1] *= target_w
      boxes[:, 2] *= target_h
      boxes[:, 3] *= target_w

    (top, left, bottom, right), (new_h, new_w) = _ComputerVisionMixin.center_image_coordinates(
      src_h=src_h,
      src_w=src_w,
      target_h=target_h,
      target_w=target_w
    )

    # eliminate centering
    boxes[:, 0] = boxes[:, 0] - top
    boxes[:, 1] = boxes[:, 1] - left
    boxes[:, 2] = boxes[:, 2] - top
    boxes[:, 3] = boxes[:, 3] - left

    # translate to original image scale
    boxes[:, 0] = boxes[:, 0] / new_h * src_h
    boxes[:, 1] = boxes[:, 1] / new_w * src_w
    boxes[:, 2] = boxes[:, 2] / new_h * src_h
    boxes[:, 3] = boxes[:, 3] / new_w * src_w

    # clipping between [0: max]
    boxes = boxes.astype(np.int32)
    boxes[:, 0] = np.maximum(0, boxes[:, 0])
    boxes[:, 1] = np.maximum(0, boxes[:, 1])
    boxes[:, 2] = np.minimum(src_h, boxes[:, 2])
    boxes[:, 3] = np.minimum(src_w, boxes[:, 3])
    return boxes

  @staticmethod
  def dir_visual_image_dedup(path_dir, magnify_image=None):
    """
    This method reads a folder, display every image in that folder al let's you
    decide where to move a specific image.
    Use:
      - s for skip
      - b for bad
      - g for good
      - q for quit
    """

    def close():
      cv2.waitKey(1)
      cv2.destroyAllWindows()
      for _ in range(5):
        cv2.waitKey(1)
      # endfor

    assert os.path.exists(path_dir)
    files = [x for x in os.listdir(path_dir) if x.endswith('.png') or x.endswith('.jpg')]
    path_good = os.path.join(path_dir, 'good')
    path_bad = os.path.join(path_dir, 'bad')
    for file in files:
      img_path = os.path.join(path_dir, file)
      img = cv2.imread(img_path)
      if magnify_image and magnify_image > 0:
        img = cv2.resize(img, (img.shape[1] * magnify_image, img.shape[0] * magnify_image))
      # endif
      cv2.imshow('Img', img)
      key = cv2.waitKey(0)
      if key == ord('b'):
        os.makedirs(path_bad, exist_ok=True)
        copyfile(img_path, os.path.join(path_bad, file))
      elif key == ord('g'):
        os.makedirs(path_good, exist_ok=True)
        copyfile(img_path, os.path.join(path_good, file))
      elif key == ord('q'):
        close()
        break
      elif key == ord('s'):
        pass
      # endif
    # endfor
    close()

  @staticmethod
  def to_rgb(image):
    return cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

  @staticmethod
  def change_brightness(img, delta):
    hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h, s, v = cv2.split(hsv)
    lim = 255 - delta
    v[v > lim] = 255
    v[v <= lim] += delta
    final_hsv = cv2.merge((h, s, v))
    img = cv2.cvtColor(final_hsv, cv2.COLOR_HSV2BGR)
    return img

  @staticmethod
  def rotate(image, angle, center=None, scale=1.0):
    (h, w) = image.shape[:2]
    if center is None:
      center = (w / 2, h / 2)
    # Perform the rotation
    M = cv2.getRotationMatrix2D(center, angle, scale)
    rotated = cv2.warpAffine(image, M, (w, h))
    return rotated

  @staticmethod
  def np_image_to_base64(np_image, ENCODING='utf-8', quality=95, max_height=None):
    from PIL import Image
    image = Image.fromarray(np_image)
    w, h = image.size
    if max_height is not None and h > max_height:
      ratio = max_height / h
      new_w, new_h = int(ratio * w), max_height
      image = image.resize((new_w, new_h), reducing_gap=1)
    buffered = BytesIO()
    if quality < 95:
      image.save(buffered, format='JPEG', quality=quality, optimize=True)
    else:
      image.save(buffered, format='JPEG')
    img_base64 = base64.b64encode(buffered.getvalue())
    img_str = img_base64.decode(ENCODING)
    return img_str

  @staticmethod
  def base64_to_np_image(base64_img):
    from PIL import Image
    base64_decoded = base64.b64decode(base64_img)
    image = Image.open(io.BytesIO(base64_decoded))
    np_image = np.array(image)
    return np_image

  @staticmethod
  def plt_to_base64(plt, close=True):
    figfile = BytesIO()
    plt.savefig(figfile, format='JPEG')
    figfile.seek(0)
    base64_bytes = base64.b64encode(figfile.getvalue())
    base64_string = base64_bytes.decode('utf-8')
    if close:
      plt.close()
    return base64_string
  
  @staticmethod
  def plt_to_np(plt, close=True, axis=False):
    # TODO: this method IS NOT thread-safe - needs revision
    if not axis:
      plt.axis('off')
    fig = plt.gcf()
    fig.canvas.draw()
    data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    w, h = fig.canvas.get_width_height()
    try:
      np_img = data.reshape((int(h), int(w), -1))
    except Exception as e:
      print(e)
      np_img = None
    if close:
      plt.close()
    return np_img    

  def time_model_forward(self, model, img_height, img_width, nr_inputs=1, lst_range=list(range(2, 33, 2))):
    from tqdm import tqdm
    lst_batches = list(range(2, 33, 2))
    self.P('Testing model {} forward time for {} iterations with batches: {}'.format(
      model.name, len(lst_batches), lst_batches)
    )
    for batch in tqdm(lst_batches):
      for i in range(2):
        data = nr_inputs * [np.random.uniform(size=(batch, img_height, img_width, 3))]
        start = tm()
        model.predict(data)
        stop = tm()
        if i == 1:
          self.P(' Forward time for bs {}: {:.3f}s. Time per obs: {:.3f}s'.format(
            batch, stop - start, (stop - start) / batch)
          )
        # endif
      # endfor
    # endfor

  @staticmethod
  def destroy_all_windows():
    cv2.waitKey(1)
    for i in range(5):
      cv2.waitKey(1)
      cv2.destroyAllWindows()
      cv2.waitKey(1)
    # endfor

  def describe_np_tensor(self, np_v):
    """
    Will take a numpy vector and will print statistics about that vector:
      - shape
      - min
      - max
      - std
      - median
      - percentiles (25 - 100)

    Parameters
    ----------
    np_v : np.ndarray
      Tensor of nD

    Returns
    -------
    None.
    """
    import pandas as pd
    assert isinstance(np_v, np.ndarray)
    lst = []
    lst.append(('Shape', np_v.shape))
    lst.append(('Dtype', np_v.dtype))
    lst.append(('Min', np.min(np_v)))
    lst.append(('Max', np.max(np_v)))
    lst.append(('Std', np.std(np_v)))
    lst.append(('Median', np.median(np_v)))
    for perc in [25, 50, 75, 100]:
      lst.append((perc, np.percentile(a=np_v, q=perc)))
    self.p('Tensor description \n{}'.format(pd.DataFrame(lst, columns=['IND', 'VAL'])))
    return

  def class_distrib(self, v):
    """
    Simple function to print the class distribution. Usefull especially for supervised classification tasks

    Parameters
    ----------
    v : list or ndarray
      Target tensor used in supervised classification problem.

    Returns
    -------
    None.

    """
    import pandas as pd
    assert isinstance(v, list) or isinstance(v, np.ndarray)
    unique, counts = np.unique(v, return_counts=True)
    self.p('Class distribution \n{}'.format(pd.DataFrame({'CLASS': unique, 'NR': counts})))
    return

  def cv_check_np_tensors(self, X, y, impose_norm=True, task='binary'):
    """
    Common pitfalls checks for input tensors of a supervised computer vision task.

    Parameters
    ----------
    X : np.ndarray
      Input tensor of a supervised computer vision task.
    y : np.ndarray
      Target tensor of a supervised computer vision task.
    impose_norm : boolean, optional
      Check if input tensor is normed. The default is True.

    Returns
    -------
    None.

    """
    assert task in ['binary', 'multiclass']
    self.p('Starting tensor sanity check')
    # check tensor lenghts are equal
    assert X.shape[0] == y.shape[0]

    # check if target is of shape (N, M)
    assert len(y.shape) == 2
    if task == 'binary': assert y.shape[-1] == 1
    if task == 'multiclass': self.p('[Info] Please ensure you have the right number of classes')

    # check if source is normed, only if impose_norm==True
    if impose_norm:
      assert X.max() <= 1.0, 'Your input tensor is not normed.'
    else:
      self.p('[Warning] Your input tensor is not normed.', color='yellow')

    # check datatypes
    if impose_norm:
      assert X.dtype == 'float32', 'Your input tensor should be of type np.float32'
    else:
      assert X.dtype == 'uint8', 'You didn\'t normed your input tensor, you should convert it to np.uint8'
    assert y.dtype == 'uint8', 'Your target tensor should be of type np.uint8'

    self.p('[Warning] RGB/BGR Please ensure that all your images are RGB', color='yellow')
    self.p('Done tensor sanity check')
    return
  
  @staticmethod
  def image_resize(image, width=None, height=None, inter=None):
    if inter is None:
      inter = cv2.INTER_AREA
    # initialize the dimensions of the image to be resized and
    # grab the image size
    dim = None
    (h, w) = image.shape[:2]

    # if both the width and height are None, then return the
    # original image
    if width is None and height is None:
      return image

    # check to see if the width is None
    if width is None:
      # calculate the ratio of the height and construct the
      # dimensions
      r = height / float(h)
      dim = (int(w * r), height)

    # otherwise, the height is None
    else:
      # calculate the ratio of the width and construct the
      # dimensions
      r = width / float(w)
      dim = (width, int(h * r))

    # resize the image
    resized = cv2.resize(image, dim, interpolation = inter)

    # return the resized image
    return resized